#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import dns as dns_apidef
from neutron_lib.api.definitions import dns_domain_keywords
from neutron_lib.api.definitions import dns_domain_ports as ports_apidef
from neutron_lib.api.definitions import subnet_dns_publish_fixed_ip as sn_dns
from neutron_lib import constants as lib_const
from oslo_log import log as logging

from neutron.plugins.ml2.extensions import subnet_dns_publish_fixed_ip

LOG = logging.getLogger(__name__)


class DnsDomainKeywordsExtensionDriver(
        subnet_dns_publish_fixed_ip.SubnetDNSPublishFixedIPExtensionDriver):

    _supported_extension_aliases = [dns_apidef.ALIAS,
                                    ports_apidef.ALIAS,
                                    sn_dns.ALIAS,
                                    dns_domain_keywords.ALIAS]

    def initialize(self):
        LOG.info("DnsDomainKeywordsExtensionDriver initialization complete")

    @staticmethod
    def _parse_dns_domain(plugin_context, domain):
        for keyword in lib_const.DNS_LABEL_KEYWORDS:
            keyword_value = getattr(plugin_context, keyword, None)
            if keyword_value is not None:
                domain = domain.replace('<' + keyword + '>', keyword_value)
            else:
                LOG.warning("Keyword <%s> does not have value in current "
                            "context and it will not be replaced in the "
                            "domain %s", keyword, domain)
        return domain



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Initial operations for dvr

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'dvr_host_macs',
        sa.Column('host', sa.String(length=255), nullable=False),
        sa.Column('mac_address', sa.String(length=32),
                  nullable=False, unique=True),
        sa.PrimaryKeyConstraint('host')
    )
    op.create_table(
        'ml2_dvr_port_bindings',
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=False),
        sa.Column('router_id', sa.String(length=36), nullable=True),
        sa.Column('vif_type', sa.String(length=64), nullable=False),
        sa.Column('vif_details', sa.String(length=4095),
                  nullable=False, server_default=''),
        sa.Column('vnic_type', sa.String(length=64),
                  nullable=False, server_default='normal'),
        sa.Column('profile', sa.String(length=4095),
                  nullable=False, server_default=''),
        sa.Column(u'status', sa.String(16), nullable=False),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id', 'host')
    )
    op.create_table(
        'csnat_l3_agent_bindings',
        sa.Column('router_id', sa.String(length=36), nullable=False),
        sa.Column('l3_agent_id', sa.String(length=36), nullable=False),
        sa.Column('host_id', sa.String(length=255), nullable=True),
        sa.Column('csnat_gw_port_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['l3_agent_id'], ['agents.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['csnat_gw_port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('router_id', 'l3_agent_id')
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

from neutron.db import migration

"""add_rbac_support_for_address_group

Revision ID: 6135a7bd4425
Revises: 1e0744e4ffea
Create Date: 2021-01-22 11:24:07.435031

"""

# revision identifiers, used by Alembic.
revision = '6135a7bd4425'
down_revision = '1e0744e4ffea'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.WALLABY]


def upgrade():
    op.create_table(
        'addressgrouprbacs', sa.MetaData(),
        sa.Column('project_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.Column('id', sa.String(length=36), nullable=False,
                  primary_key=True),
        sa.Column('target_tenant', sa.String(length=255), nullable=False),
        sa.Column('action', sa.String(length=255), nullable=False),
        sa.Column('object_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['object_id'], ['address_groups.id'],
                                ondelete='CASCADE'),
        sa.UniqueConstraint('target_tenant', 'object_id', 'action',
                            name='uniq_address_groups_rbacs0'
                                 'target_tenant0object_id0action')
    )



# Copyright (c) 2017 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import data_plane_status as dps_lib
from neutron_lib.plugins.ml2 import api
from oslo_log import log as logging

from neutron.db import data_plane_status_db as dps_db

LOG = logging.getLogger(__name__)


class DataPlaneStatusExtensionDriver(api.ExtensionDriver,
                                     dps_db.DataPlaneStatusMixin):
    _supported_extension_alias = 'data-plane-status'

    def initialize(self):
        LOG.info("DataPlaneStatusExtensionDriver initialization complete")

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_update_port(self, plugin_context, data, result):
        if dps_lib.DATA_PLANE_STATUS in data:
            self._process_update_port_data_plane_status(plugin_context,
                                                        data, result)

    def extend_port_dict(self, session, db_data, result):
        self._extend_port_data_plane_status(result, db_data)



# Copyright (c) 2013 NEC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import random
import unittest

import fixtures
import netaddr
from neutron_lib.callbacks import priority_group
from neutron_lib import constants
from neutron_lib.services.logapi import constants as log_const
from neutron_lib.utils import helpers
from neutron_lib.utils import net
from oslo_utils import netutils
from oslo_utils import timeutils


# NOTE(ykarel): from neutron-lib 3.9.0, cancellable flag was added
# test the existence of the is_cancellable_event function to check if
# cancellable flag is supported or not. This compatibility check can
# be removed once neutron-lib >= 3.9.0 in requirements.txt.
_CANCELLABLE_FLAG_SUPPORTED = True
try:
    from neutron_lib.callbacks.events import is_cancellable_event  # noqa
except ImportError:
    _CANCELLABLE_FLAG_SUPPORTED = False

LAST_RANDOM_PORT_RANGE_GENERATED = 1


class SafeCleanupFixture(fixtures.Fixture):
    """Catch errors in daughter fixture cleanup."""

    def __init__(self, fixture):
        self.fixture = fixture

    def _setUp(self):

        def cleanUp():
            try:
                self.fixture.cleanUp()
            except Exception:
                pass

        self.fixture.setUp()
        self.addCleanup(cleanUp)


def setup_mock_calls(mocked_call, expected_calls_and_values):
    """A convenient method to setup a sequence of mock calls.

    expected_calls_and_values is a list of (expected_call, return_value):

        expected_calls_and_values = [
            (mock.call(["ovs-vsctl", self.TO, '--', "--may-exist", "add-port",
                        self.BR_NAME, pname]),
             None),
            (mock.call(["ovs-vsctl", self.TO, "set", "Interface",
                        pname, "type=gre"]),
             None),
            ....
        ]

    * expected_call should be mock.call(expected_arg, ....)
    * return_value is passed to side_effect of a mocked call.
      A return value or an exception can be specified.
    """
    return_values = [call[1] for call in expected_calls_and_values]
    mocked_call.side_effect = return_values


def verify_mock_calls(mocked_call, expected_calls_and_values,
                      any_order=False):
    """A convenient method to setup a sequence of mock calls.

    expected_calls_and_values is a list of (expected_call, return_value):

        expected_calls_and_values = [
            (mock.call(["ovs-vsctl", self.TO, '--', "--may-exist", "add-port",
                        self.BR_NAME, pname]),
             None),
            (mock.call(["ovs-vsctl", self.TO, "set", "Interface",
                        pname, "type=gre"]),
             None),
            ....
        ]

    * expected_call should be mock.call(expected_arg, ....)
    * return_value is passed to side_effect of a mocked call.
      A return value or an exception can be specified.
    """
    expected_calls = [call[0] for call in expected_calls_and_values]
    mocked_call.assert_has_calls(expected_calls, any_order=any_order)


def _make_magic_method(method_mock):
    # NOTE(yamahata): new environment needs to be created to keep actual
    # method_mock for each callables.
    def __call__(*args, **kwargs):
        value_mock = method_mock._orig___call__(*args, **kwargs)
        value_mock.__json__ = lambda: {}
        return value_mock

    def _get_child_mock(**kwargs):
        value_mock = method_mock._orig__get_child_mock(**kwargs)
        value_mock.__json__ = lambda: {}
        return value_mock

    return __call__, _get_child_mock


def make_mock_plugin_json_encodable(plugin_instance_mock):
    # NOTE(yamahata): Make return value of plugin method json encodable
    # e.g. the return value of plugin_instance.create_network() needs
    # to be json encodable
    # plugin instance      -> method    -> return value
    # Mock                    MagicMock    Mock
    # plugin_instance_mock    method_mock  value_mock
    #
    # From v1.3 of pecan, pecan.jsonify uses json.Encoder unconditionally.
    # pecan v1.2 uses simplejson.Encoder which accidentally encodes
    # Mock as {} due to check of '_asdict' attributes.
    # pecan.jsonify uses __json__ magic method for encoding when
    # it's defined, so add __json__ method to return {}
    for method_mock in plugin_instance_mock._mock_children.values():
        if not callable(method_mock):
            continue

        method_mock._orig___call__ = method_mock.__call__
        method_mock._orig__get_child_mock = method_mock._get_child_mock
        __call__, _get_child_mock = _make_magic_method(method_mock)
        method_mock.__call__ = __call__
        method_mock._get_child_mock = _get_child_mock


def get_subscribe_args(*args):
    args = list(args)  # don't modify original list
    args.append(priority_group.PRIORITY_DEFAULT)
    # NOTE(ykarel): from neutron-lib 3.9.0, cancellable flag was added.
    # old signature: (callback, resource, event, priority=PRIORITY_DEFAULT)
    # new signature: (callback, resource, event, priority=PRIORITY_DEFAULT,
    # cancellable=False)
    if len(args) == 4 and _CANCELLABLE_FLAG_SUPPORTED:
        args.append(False)
    return args


def fail(msg=None):
    """Fail immediately, with the given message.

    This method is equivalent to TestCase.fail without requiring a
    testcase instance (usefully for reducing coupling).
    """
    raise unittest.TestCase.failureException(msg)


def get_random_string_list(i=3, n=5):
    return [helpers.get_random_string(n) for _ in range(0, i)]


def get_random_boolean():
    return bool(random.getrandbits(1))


def get_random_datetime(start_time=None,
                        end_time=None):
    start_time = start_time or timeutils.utcnow()
    end_time = end_time or (start_time + datetime.timedelta(days=1))
    # calculate the seconds difference between start and end time
    delta_seconds_difference = int(timeutils.delta_seconds(start_time,
                                                           end_time))
    # get a random time_delta_seconds between 0 and
    # delta_seconds_difference
    random_time_delta = random.randint(0, delta_seconds_difference)
    # generate a random datetime between start and end time
    return start_time + datetime.timedelta(seconds=random_time_delta)


def get_random_integer(range_begin=0, range_end=1000):
    return random.randint(range_begin, range_end)


def get_random_prefixlen(version=4):
    maxlen = constants.IPv4_BITS
    if version == 6:
        maxlen = constants.IPv6_BITS
    return random.randint(0, maxlen)


def get_random_port(start=constants.PORT_RANGE_MIN):
    global LAST_RANDOM_PORT_RANGE_GENERATED
    LAST_RANDOM_PORT_RANGE_GENERATED = random.randint(
        start, constants.PORT_RANGE_MAX)
    return LAST_RANDOM_PORT_RANGE_GENERATED


def get_random_vlan():
    return random.randint(constants.MIN_VLAN_TAG, constants.MAX_VLAN_TAG)


def get_random_ip_version():
    return random.choice(constants.IP_ALLOWED_VERSIONS)


def get_random_ip_address(version=4):
    if version == 4:
        ip_string = '10.%d.%d.%d' % (random.randint(3, 254),
                                     random.randint(3, 254),
                                     random.randint(3, 254))
        return netaddr.IPAddress(ip_string)
    else:
        ip = netutils.get_ipv6_addr_by_EUI64(
            '2001:db8::/64',
            net.get_random_mac(['fe', '16', '3e', '00', '00', '00'])
        )
        return ip


def get_random_router_status():
    return random.choice(constants.VALID_ROUTER_STATUS)


def get_random_floatingip_status():
    return random.choice(constants.VALID_FLOATINGIP_STATUS)


def get_random_flow_direction():
    return random.choice(constants.VALID_DIRECTIONS)


def get_random_flow_direction_or_any():
    return random.choice(constants.VALID_DIRECTIONS_AND_ANY)


def get_random_ha_states():
    return random.choice(constants.VALID_HA_STATES)


def get_random_ether_type():
    return random.choice(constants.VALID_ETHERTYPES)


def get_random_ipam_status():
    return random.choice(constants.VALID_IPAM_ALLOCATION_STATUSES)


def get_random_ip_protocol():
    return random.choice(list(constants.IP_PROTOCOL_MAP.keys()))


def get_random_port_binding_statuses():
    return random.choice(constants.PORT_BINDING_STATUSES)


def get_random_network_segment_range_network_type():
    return random.choice([constants.TYPE_VLAN,
                          constants.TYPE_VXLAN,
                          constants.TYPE_GRE,
                          constants.TYPE_GENEVE])


def get_random_ipv6_mode():
    return random.choice(constants.IPV6_MODES)


def get_random_security_event():
    return random.choice(log_const.LOG_EVENTS)


def get_random_port_numa_affinity_policy():
    return random.choice(constants.PORT_NUMA_POLICIES)


def get_random_port_hardware_offload_type():
    return random.choice(constants.VALID_HWOL_TYPES)



# Copyright (c) 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import random

from neutron_lib.agent import topics
from neutron_lib.api import extensions
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging

from neutron.common import utils


# Priorities - lower value is higher priority
PRIORITY_NETWORK_CREATE = 0
PRIORITY_NETWORK_UPDATE = 1
PRIORITY_NETWORK_DELETE = 2
PRIORITY_SUBNET_UPDATE = 3
PRIORITY_SUBNET_DELETE = 4
# In order to improve port dhcp provisioning when nova concurrently create
# multiple vms, I classify the port_create_end message to two levels, the
# high-level message only cast to one agent, the low-level message cast to all
# other agent. In this way, When there are a large number of ports that need to
# be processed, we can dispatch the high priority message of port to different
# agent, so that the processed port will not block other port's processing in
# other dhcp agents.
PRIORITY_PORT_CREATE_HIGH = 5
PRIORITY_PORT_CREATE_LOW = 6
PRIORITY_PORT_UPDATE = 6
PRIORITY_PORT_DELETE = 6

METHOD_PRIORITY_MAP = {
    'network_create_end': PRIORITY_NETWORK_CREATE,
    'network_update_end': PRIORITY_NETWORK_UPDATE,
    'network_delete_end': PRIORITY_NETWORK_DELETE,
    'subnet_create_end': PRIORITY_SUBNET_UPDATE,
    'subnet_update_end': PRIORITY_SUBNET_UPDATE,
    'subnet_delete_end': PRIORITY_SUBNET_DELETE,
    'port_create_end': PRIORITY_PORT_CREATE_LOW,
    'port_update_end': PRIORITY_PORT_UPDATE,
    'port_delete_end': PRIORITY_PORT_DELETE
}


LOG = logging.getLogger(__name__)


class DhcpAgentNotifyAPI(object):
    """API for plugin to notify DHCP agent.

    This class implements the client side of an rpc interface.  The server side
    is neutron.agent.dhcp.agent.DhcpAgent.  For more information about changing
    rpc interfaces, please see doc/source/contributor/internals/rpc_api.rst.
    """
    # It seems dhcp agent does not support bulk operation
    VALID_RESOURCES = ['network', 'subnet', 'port']
    VALID_METHOD_NAMES = ['network.create.end',
                          'network.update.end',
                          'network.delete.end',
                          'subnet.create.end',
                          'subnet.update.end',
                          'subnet.delete.end',
                          'port.create.end',
                          'port.update.end',
                          'port.delete.end']

    def __init__(self, topic=topics.DHCP_AGENT, plugin=None):
        self._unsubscribed_resources = []
        self._plugin = plugin
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)
        if not cfg.CONF.dhcp_agent_notification:
            return
        # register callbacks for router interface changes
        registry.subscribe(self._after_router_interface_created,
                           resources.ROUTER_INTERFACE, events.AFTER_CREATE)
        registry.subscribe(self._after_router_interface_deleted,
                           resources.ROUTER_INTERFACE, events.AFTER_DELETE)
        # register callbacks for events pertaining resources affecting DHCP
        callback_resources = (
            resources.NETWORK,
            resources.NETWORKS,
            resources.PORT,
            resources.PORTS,
            resources.SUBNET,
            resources.SUBNETS,
        )
        for resource in callback_resources:
            registry.subscribe(self._send_dhcp_notification,
                               resource, events.BEFORE_RESPONSE)
        self.uses_native_notifications = {}
        for resource in (resources.NETWORK, resources.PORT, resources.SUBNET):
            self.uses_native_notifications[resource] = {'create': False,
                                                        'update': False,
                                                        'delete': False}
            registry.subscribe(self._native_event_send_dhcp_notification,
                               resource, events.AFTER_CREATE)
            registry.subscribe(self._native_event_send_dhcp_notification,
                               resource, events.AFTER_UPDATE)
            registry.subscribe(self._native_event_send_dhcp_notification,
                               resource, events.AFTER_DELETE)

    @property
    def plugin(self):
        if self._plugin is None:
            self._plugin = directory.get_plugin()
        return self._plugin

    def _schedule_network(self, context, network, existing_agents):
        """Schedule the network to new agents

        :return: all agents associated with the network
        """
        new_agents = self.plugin.schedule_network(context, network) or []
        if new_agents:
            for agent in new_agents:
                self._cast_message(
                    context, 'network_create_end',
                    {'network': {'id': network['id']},
                     'priority': PRIORITY_NETWORK_CREATE}, agent['host'])
        elif not existing_agents:
            LOG.warning('Unable to schedule network %s: no agents '
                        'available; will retry on subsequent port '
                        'and subnet creation events.',
                        network['id'])
        return new_agents + existing_agents

    def _get_enabled_agents(
            self, context, network_id, network, agents, method, payload):
        """Get the list of agents who can provide services."""
        if not agents:
            return []
        enabled_agents = agents
        if not cfg.CONF.enable_services_on_agents_with_admin_state_down:
            enabled_agents = [x for x in agents if x.admin_state_up]
        active_agents = [x for x in agents if x.is_active]
        len_enabled_agents = len(enabled_agents)
        len_active_agents = len(active_agents)
        if len_active_agents < len_enabled_agents:
            LOG.warning("Only %(active)d of %(total)d DHCP agents "
                        "associated with network '%(net_id)s' "
                        "are marked as active, so notifications "
                        "may be sent to inactive agents.",
                        {'active': len_active_agents,
                         'total': len_enabled_agents,
                         'net_id': network_id})
        if not enabled_agents:
            num_ports = self.plugin.get_ports_count(
                context, {'network_id': [network_id]})
            if not network:
                admin_ctx = context if context.is_admin else context.elevated()
                network = self.plugin.get_network(admin_ctx, network_id)
            notification_required = (
                num_ports > 0 and len(network['subnets']) >= 1)
            if notification_required:
                LOG.error("Will not send event %(method)s for network "
                          "%(net_id)s: no agent available. Payload: "
                          "%(payload)s",
                          {'method': method,
                           'net_id': network_id,
                           'payload': payload})
        return enabled_agents

    def _is_reserved_dhcp_port(self, port):
        return port.get('device_id') == constants.DEVICE_ID_RESERVED_DHCP_PORT

    @utils.disable_notifications
    def _notify_agents(
            self, context, method, payload, network_id, network=None):
        """Notify all the agents that are hosting the network."""
        payload['priority'] = METHOD_PRIORITY_MAP.get(method)
        # fanout is required as we do not know who is "listening"
        no_agents = not extensions.is_extension_supported(
            self.plugin, constants.DHCP_AGENT_SCHEDULER_EXT_ALIAS)
        fanout_required = method == 'network_delete_end' or no_agents

        # we do nothing on network creation because we want to give the
        # admin the chance to associate an agent to the network manually
        cast_required = method != 'network_create_end'

        if fanout_required:
            self._fanout_message(context, method, payload)
        elif cast_required:
            candidate_hosts = None
            if 'subnet' in payload and payload['subnet'].get('segment_id'):
                # if segment_id exists then the segment service plugin
                # must be loaded
                segment_plugin = directory.get_plugin('segments')
                segment = segment_plugin.get_segment(
                    context, payload['subnet']['segment_id'])
                candidate_hosts = segment['hosts']

            agents = self.plugin.get_dhcp_agents_hosting_networks(
                context, [network_id], hosts=candidate_hosts)
            # schedule the network first, if needed
            schedule_required = (
                method == 'subnet_create_end' or
                method == 'port_create_end' and
                not self._is_reserved_dhcp_port(payload['port']))
            if schedule_required:
                admin_ctx = context if context.is_admin else context.elevated()
                network = network or self.plugin.get_network(
                    admin_ctx, network_id)
                if candidate_hosts:
                    network['candidate_hosts'] = candidate_hosts
                agents = self._schedule_network(admin_ctx, network, agents)
            if not agents:
                LOG.debug("Network %s is not hosted by any dhcp agent",
                          network_id)
                return
            enabled_agents = self._get_enabled_agents(
                context, network_id, network, agents, method, payload)

            if method == 'port_create_end' and enabled_agents:
                high_agent = enabled_agents.pop(
                    random.randint(0, len(enabled_agents) - 1))
                self._notify_high_priority_agent(
                    context, copy.deepcopy(payload), high_agent)
            for agent in enabled_agents:
                self._cast_message(
                    context, method, payload, agent.host, agent.topic)

    def _notify_high_priority_agent(self, context, payload, agent):
        payload['priority'] = PRIORITY_PORT_CREATE_HIGH
        self._cast_message(context, "port_create_end",
                           payload, agent.host, agent.topic)

    @utils.disable_notifications
    def _cast_message(self, context, method, payload, host,
                      topic=topics.DHCP_AGENT):
        """Cast the payload to the dhcp agent running on the host."""
        cctxt = self.client.prepare(topic=topic, server=host)
        cctxt.cast(context, method, payload=payload)

    @utils.disable_notifications
    def _fanout_message(self, context, method, payload):
        """Fanout the payload to all dhcp agents."""
        cctxt = self.client.prepare(fanout=True)
        cctxt.cast(context, method, payload=payload)

    def network_removed_from_agent(self, context, network_id, host):
        self._cast_message(context, 'network_delete_end',
                           {'network_id': network_id,
                            'priority': PRIORITY_NETWORK_DELETE}, host)

    def network_added_to_agent(self, context, network_id, host):
        self._cast_message(context, 'network_create_end',
                           {'network': {'id': network_id},
                            'priority': PRIORITY_NETWORK_CREATE}, host)

    def agent_updated(self, context, admin_state_up, host):
        self._cast_message(context, 'agent_updated',
                           {'admin_state_up': admin_state_up}, host)

    def _after_router_interface_created(self, resource, event, trigger,
                                        payload=None):
        port = payload.metadata.get('port')
        self._notify_agents(payload.context, 'port_create_end',
                            {'port': port},
                            port['network_id'])

    def _after_router_interface_deleted(self, resource, event, trigger,
                                        payload=None):
        port = payload.metadata.get('port')
        self._notify_agents(payload.context, 'port_delete_end',
                            {'port_id': port['id'],
                             'network_id': port['network_id']},
                            port['network_id'])

    def _native_event_send_dhcp_notification(self, resource, event, trigger,
                                             payload):
        action = event.replace('after_', '')
        # we unsubscribe the _send_dhcp_notification method now that we know
        # the loaded core plugin emits native resource events
        if resource not in self._unsubscribed_resources:
            self.uses_native_notifications[resource][action] = True
            if all(self.uses_native_notifications[resource].values()):
                # only unsubscribe the API level listener if we are
                # receiving all event types for this resource
                self._unsubscribed_resources.append(resource)
                registry.unsubscribe_by_resource(self._send_dhcp_notification,
                                                 resource)
        method_name = '.'.join((resource, action, 'end'))
        data = {resource: payload.latest_state}
        if resource == resources.PORT and event == events.AFTER_UPDATE:
            if not self._notification_is_needed(payload.states[0],
                                                payload.latest_state):
                # don't waste time updating the DHCP agent for status updates
                return
        self.notify(payload.context, data, method_name)

    def _notification_is_needed(self, orig, new):
        # notification to the DHCP agent should be send only if DHCP related
        # attributes of the port were changed, like: fixed_ips, mac_address,
        # extra_dhcp_opts, dns_name, dns_assignment or dns_domain.
        # Otherwise there is no need to waste DHCP agent's time for doing
        # updates.
        if not orig or not new:
            return True
        return any(orig.get(k) != new.get(k) for k in (
            'fixed_ips', 'mac_address', 'extra_dhcp_opts',
            'dns_name', 'dns_assignment', 'dns_domain'))

    def _send_dhcp_notification(self, resource, event, trigger, payload=None):
        action = payload.action.split('_')[0]
        if (resource in self.uses_native_notifications and
                self.uses_native_notifications[resource][action]):
            return
        data = payload.latest_state
        if payload.collection_name and payload.collection_name in data:
            for body in data[payload.collection_name]:
                item = {resource: body}
                self.notify(payload.context, item, payload.method_name)
        else:
            self.notify(payload.context, data, payload.method_name)

    def notify(self, context, data, method_name):
        # data is {'key' : 'value'} with only one key
        if method_name not in self.VALID_METHOD_NAMES:
            return
        obj_type = list(data.keys())[0]
        if obj_type not in self.VALID_RESOURCES:
            return
        obj_value = data[obj_type]
        network_id = None
        if obj_type == 'network' and 'id' in obj_value:
            network_id = obj_value['id']
        elif obj_type in ['port', 'subnet'] and 'network_id' in obj_value:
            network_id = obj_value['network_id']
        if not network_id:
            return
        method_name = method_name.replace(".", "_")
        if method_name.endswith("_delete_end"):
            if 'id' in obj_value:
                payload = {obj_type + '_id': obj_value['id']}
                if obj_type != 'network':
                    payload['network_id'] = network_id
                if obj_type == 'port':
                    payload['fixed_ips'] = obj_value['fixed_ips']
                self._notify_agents(context, method_name, payload, network_id,
                                    obj_value.get('network'))
        else:
            self._notify_agents(context, method_name, data, network_id,
                                obj_value.get('network'))



# Copyright 2014 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from oslo_config import cfg
from oslo_utils import uuidutils
import pyroute2

from neutron.agent.l2 import l2_agent_extensions_manager as l2_ext_manager
from neutron.agent import rpc as agent_rpc
from neutron.plugins.ml2.drivers.mech_sriov.agent.common import config  # noqa
from neutron.plugins.ml2.drivers.mech_sriov.agent import sriov_nic_agent
from neutron.privileged.agent.linux import ip_lib as priv_ip_lib
from neutron.tests import base

DEVICE_MAC = '11:22:33:44:55:66'
PCI_SLOT = "0000:06:00.1"
DEV1 = agent_rpc.DeviceInfo('mac1', 'pci_slot1')
DEV2 = agent_rpc.DeviceInfo('mac2', 'pci_slot2')
DEV3 = agent_rpc.DeviceInfo('mac3', 'pci_slot3')
DEV4 = agent_rpc.DeviceInfo('mac4', 'pci_slot4')
RPC_DEV1 = {'device': DEV1.mac,
            'port_id': 'port123',
            'network_id': 'net123',
            'admin_state_up': True,
            'propagate_uplink_status': False,
            'network_type': 'vlan',
            'segmentation_id': 100,
            'profile': {'pci_slot': DEV1.pci_slot},
            'physical_network': 'physnet1',
            'port_security_enabled': False}
RPC_DEV2 = {'device': DEV2.mac,
            'port_id': 'port321',
            'network_id': 'net123',
            'admin_state_up': True,
            'propagate_uplink_status': False,
            'network_type': 'vlan',
            'segmentation_id': 100,
            'profile': {'pci_slot': DEV2.pci_slot},
            'physical_network': 'physnet1',
            'port_security_enabled': False}


class TestSriovAgent(base.BaseTestCase):
    def setUp(self):
        super(TestSriovAgent, self).setUp()
        # disable setting up periodic state reporting
        cfg.CONF.set_override('report_interval', 0, 'AGENT')
        cfg.CONF.set_default('firewall_driver',
                             'neutron.agent.firewall.NoopFirewallDriver',
                             group='SECURITYGROUP')
        cfg.CONF.set_default('enable_security_group',
                             False,
                             group='SECURITYGROUP')

        class MockFixedIntervalLoopingCall(object):
            def __init__(self, f):
                self.f = f

            def start(self, interval=0):
                self.f()

        mock.patch('oslo_service.loopingcall.'
                   'FixedIntervalLoopingCall',
                   new=MockFixedIntervalLoopingCall)

        self.agent = sriov_nic_agent.SriovNicSwitchAgent({}, {}, 0, {}, {}, {})

    @mock.patch("neutron.plugins.ml2.drivers.mech_sriov.agent.eswitch_manager"
               ".ESwitchManager.get_assigned_devices_info", return_value=set())
    @mock.patch.object(agent_rpc.PluginReportStateAPI, 'report_state')
    def test_cached_device_count_report_state(self, report_state, get_dev):
        self.agent._report_state()
        agent_conf = self.agent.agent_state['configurations']
        # ensure devices aren't calculated until first scan_devices call
        self.assertNotIn('devices', agent_conf)
        self.agent.scan_devices(set(), set())
        self.assertEqual(0, agent_conf['devices'])
        # ensure report_state doesn't call get_dev
        get_dev.reset_mock()
        get_dev.return_value = set(['dev1', 'dev2'])
        self.agent._report_state()
        self.assertEqual(0, agent_conf['devices'])
        # after a device scan, conf should bump to 2
        self.agent.scan_devices(set(), set())
        self.assertEqual(2, agent_conf['devices'])

    def test_treat_devices_removed_with_existed_device(self, *args):
        agent = sriov_nic_agent.SriovNicSwitchAgent({}, {}, 0, {}, {}, {})
        devices = {agent_rpc.DeviceInfo(DEVICE_MAC, PCI_SLOT)}
        with mock.patch.object(agent.plugin_rpc,
                               "update_device_down") as fn_udd:
            fn_udd.return_value = {'device': DEVICE_MAC,
                                   'exists': True}
            resync = agent.treat_devices_removed(devices)
            self.assertFalse(resync)
            self.assertTrue(fn_udd.called)

    def test_treat_devices_removed_with_not_existed_device(self, *args):
        agent = sriov_nic_agent.SriovNicSwitchAgent({}, {}, 0, {}, {}, {})
        devices = {agent_rpc.DeviceInfo(DEVICE_MAC, PCI_SLOT)}
        with mock.patch.object(agent.plugin_rpc,
                               "update_device_down") as fn_udd:
            fn_udd.return_value = {'device': DEVICE_MAC,
                                   'exists': False}
            with mock.patch.object(sriov_nic_agent.LOG,
                                   'debug') as log:
                resync = agent.treat_devices_removed(devices)
                self.assertEqual(1, log.call_count)
                self.assertFalse(resync)
                self.assertTrue(fn_udd.called)

    def test_treat_devices_removed_failed(self, *args):
        agent = sriov_nic_agent.SriovNicSwitchAgent({}, {}, 0, {}, {}, {})
        devices = {agent_rpc.DeviceInfo(DEVICE_MAC, PCI_SLOT)}
        with mock.patch.object(agent.plugin_rpc,
                               "update_device_down") as fn_udd:
            fn_udd.side_effect = Exception()
            with mock.patch.object(sriov_nic_agent.LOG,
                                   'debug') as log:
                resync = agent.treat_devices_removed(devices)
                self.assertEqual(1, log.call_count)
                self.assertTrue(resync)
                self.assertTrue(fn_udd.called)

    def mock_scan_devices(self, expected, mock_current,
                          registered_devices, updated_devices):
        self.agent.eswitch_mgr = mock.Mock()
        self.agent.eswitch_mgr.get_assigned_devices_info.return_value = (
            mock_current)

        results = self.agent.scan_devices(registered_devices, updated_devices)
        self.assertEqual(expected, results)

    def test_scan_devices_returns_empty_sets(self):
        registered = set()
        updated = set()
        mock_current = set()
        expected = {'current': set(),
                    'updated': set(),
                    'added': set(),
                    'removed': set()}
        self.mock_scan_devices(expected, mock_current, registered, updated)

    def test_scan_devices_no_changes(self):
        registered = {DEV1, DEV2}
        updated = set()
        mock_current = {DEV1, DEV2}
        expected = {'current': {DEV1, DEV2},
                    'updated': set(),
                    'added': set(),
                    'removed': set()}
        self.mock_scan_devices(expected, mock_current, registered, updated)

    def test_scan_devices_new_and_removed(self):
        registered = {DEV1, DEV2}
        updated = set()
        mock_current = {DEV2, DEV3}
        expected = {'current': {DEV2, DEV3},
                    'updated': set(),
                    'added': {DEV3},
                    'removed': {DEV1}}
        self.mock_scan_devices(expected, mock_current, registered, updated)

    def test_scan_devices_updated_and_removed(self):
        registered = {DEV1, DEV2}
        # 'DEV1' is in removed and updated tuple
        updated = {DEV1}
        mock_current = {DEV2, DEV3}
        expected = {'current': {DEV2, DEV3},
                    'updated': set(),
                    'added': {DEV3},
                    'removed': {DEV1}}
        self.mock_scan_devices(expected, mock_current, registered, updated)

    def test_scan_devices_new_updates(self):
        registered = {DEV1}
        updated = {DEV2}
        mock_current = {DEV1, DEV2}
        expected = {'current': {DEV1, DEV2},
                    'updated': {DEV2},
                    'added': {DEV2},
                    'removed': set()}
        self.mock_scan_devices(expected, mock_current, registered, updated)

    def test_scan_devices_updated_missing(self):
        registered = {DEV1}
        updated = {DEV2}
        mock_current = {DEV1}
        expected = {'current': {DEV1},
                    'updated': set(),
                    'added': set(),
                    'removed': set()}
        self.mock_scan_devices(expected, mock_current, registered, updated)

    def test_process_network_devices(self):
        agent = self.agent
        device_info = {'current': set(),
                       'added': {DEV3, DEV4},
                       'updated': {DEV2, DEV3},
                       'removed': {DEV1}}
        agent.sg_agent.prepare_devices_filter = mock.Mock()
        agent.sg_agent.refresh_firewall = mock.Mock()
        agent.treat_devices_added_updated = mock.Mock(return_value=False)
        agent.treat_devices_removed = mock.Mock(return_value=False)

        agent.process_network_devices(device_info)

        agent.sg_agent.prepare_devices_filter.assert_called_with({DEV3, DEV4})
        self.assertTrue(agent.sg_agent.refresh_firewall.called)
        agent.treat_devices_added_updated.assert_called_with(
            {DEV2, DEV3, DEV4})
        agent.treat_devices_removed.assert_called_with({DEV1})

    def test_treat_devices_added_updated_sends_host(self):
        agent = self.agent
        host = 'host1'
        cfg.CONF.set_override('host', host)
        agent.plugin_rpc = mock.Mock()
        agent.plugin_rpc.get_devices_details_list.return_value = [RPC_DEV1]
        devices = {agent_rpc.DeviceInfo(DEV1.mac, DEV1.pci_slot)}
        agent.treat_devices_added_updated(devices)
        agent.plugin_rpc.get_devices_details_list.assert_called_once_with(
            agent.context, devices, agent.agent_id, host)

    def test_treat_devices_added_updated_and_removed(self):
        agent = self.agent
        agent.plugin_rpc = mock.Mock()
        agent.plugin_rpc.get_devices_details_list.return_value = [RPC_DEV1]
        agent.treat_devices_added_updated({DEV1})
        self.assertEqual({'net123': [{'port_id': 'port123', 'device': DEV1}]},
                         agent.network_ports)
        agent.plugin_rpc.get_devices_details_list.return_value = [RPC_DEV2]
        # add the second device and check the network_ports dict
        agent.treat_devices_added_updated({DEV2})
        self.assertEqual({'net123': [{'port_id': 'port123', 'device': DEV1},
                                     {'port_id': 'port321', 'device': DEV2}]},
                         agent.network_ports)
        with mock.patch.object(agent.plugin_rpc, "update_device_down"):
            agent.treat_devices_removed({DEV2})
        # remove the second device and check the network_ports dict
        self.assertEqual({'net123': [{'port_id': 'port123', 'device': DEV1}]},
                         agent.network_ports)

    def test_treat_devices_added_updated_admin_state_up_true(self):
        agent = self.agent
        agent.plugin_rpc = mock.Mock()
        agent.plugin_rpc.get_devices_details_list.return_value = [RPC_DEV1]
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = True
        agent.set_device_state = mock.Mock()
        agent.set_device_spoofcheck = mock.Mock()
        devices = {DEV1}
        resync_needed = agent.treat_devices_added_updated(devices)
        self.assertFalse(resync_needed)
        agent.eswitch_mgr.device_exists.assert_called_with(DEV1.mac,
                                                           DEV1.pci_slot)
        agent.eswitch_mgr.set_device_state.assert_called_with(
            DEV1.mac, DEV1.pci_slot, True, False)
        agent.eswitch_mgr.set_device_spoofcheck.assert_called_with(
            DEV1.mac, DEV1.pci_slot, False)
        agent.plugin_rpc.update_device_list.assert_called_once_with(
            agent.context, devices, set(), agent.agent_id, agent.conf.host)

    def test_treat_devices_added_updated_multiple_admin_state_up_true(self):
        agent = self.agent
        mock_details = [RPC_DEV1, RPC_DEV2]
        agent.plugin_rpc = mock.Mock()
        agent.plugin_rpc.get_devices_details_list.return_value = mock_details
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = True
        agent.set_device_state = mock.Mock()
        agent.set_device_spoofcheck = mock.Mock()
        devices = {DEV1, DEV2}
        resync_needed = agent.treat_devices_added_updated(devices)
        self.assertFalse(resync_needed)
        calls = [mock.call(DEV1.mac, DEV1.pci_slot),
                 mock.call(DEV2.mac, DEV2.pci_slot)]
        agent.eswitch_mgr.device_exists.assert_has_calls(calls, any_order=True)
        calls = [mock.call(DEV1.mac, DEV1.pci_slot, True, False),
                 mock.call(DEV2.mac, DEV2.pci_slot, True, False)]
        agent.eswitch_mgr.set_device_state.assert_has_calls(calls,
                                                            any_order=True)
        calls = [mock.call(DEV1.mac, DEV1.pci_slot, False),
                 mock.call(DEV2.mac, DEV2.pci_slot, False)]
        agent.eswitch_mgr.set_device_spoofcheck.assert_has_calls(
            calls, any_order=True)
        agent.plugin_rpc.update_device_list.assert_called_once_with(
            agent.context, devices, set(), agent.agent_id, agent.conf.host)

    def test_treat_devices_added_updated_multiple_admin_states(self):
        agent = self.agent
        rpc_dev2 = copy.deepcopy(RPC_DEV2)
        rpc_dev2['admin_state_up'] = False
        mock_details = [RPC_DEV1, rpc_dev2]
        agent.plugin_rpc = mock.Mock()
        agent.plugin_rpc.get_devices_details_list.return_value = mock_details
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = True
        agent.set_device_state = mock.Mock()
        agent.set_device_spoofcheck = mock.Mock()
        devices = {DEV1, DEV2}
        resync_needed = agent.treat_devices_added_updated(devices)
        self.assertFalse(resync_needed)
        calls = [mock.call(DEV1.mac, DEV1.pci_slot),
                 mock.call(DEV2.mac, DEV2.pci_slot)]
        agent.eswitch_mgr.device_exists.assert_has_calls(calls, any_order=True)
        calls = [mock.call(DEV1.mac, DEV1.pci_slot, True, False),
                 mock.call(DEV2.mac, DEV2.pci_slot, False, False)]
        agent.eswitch_mgr.set_device_state.assert_has_calls(calls,
                                                            any_order=True)
        calls = [mock.call(DEV1.mac, DEV1.pci_slot, False),
                 mock.call(DEV2.mac, DEV2.pci_slot, False)]
        agent.eswitch_mgr.set_device_spoofcheck.assert_has_calls(
            calls, any_order=True)
        agent.plugin_rpc.update_device_list.assert_called_once_with(
            agent.context, {DEV1}, {DEV2}, agent.agent_id, agent.conf.host)

    def test_treat_device_ip_link_state_not_supported(self):
        agent = self.agent
        agent.plugin_rpc = mock.Mock()
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = True
        agent.eswitch_mgr.set_device_state.side_effect = (
            priv_ip_lib.InterfaceOperationNotSupported())

        self.assertTrue(agent.treat_device(DEV1, admin_state_up=True))

    def test_treat_device_set_device_state_exception(self):
        agent = self.agent
        agent.plugin_rpc = mock.Mock()
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = True
        agent.eswitch_mgr.set_device_state.side_effect = (
            pyroute2.NetlinkError(22))

        self.assertFalse(agent.treat_device(DEV1, admin_state_up=True))

    def test_treat_device_no_device_found(self):
        agent = self.agent
        agent.plugin_rpc = mock.Mock()
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = False

        self.assertFalse(agent.treat_device(DEV1, admin_state_up=True))

    def test_treat_devices_added_updated_admin_state_up_false(self):
        agent = self.agent
        agent.plugin_rpc = mock.Mock()
        rpc_dev1 = copy.deepcopy(RPC_DEV1)
        rpc_dev1['admin_state_up'] = False
        agent.plugin_rpc.get_devices_details_list.return_value = [rpc_dev1]
        agent.remove_port_binding = mock.Mock()
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = True
        devices = {DEV1}
        resync_needed = agent.treat_devices_added_updated(devices)
        self.assertFalse(resync_needed)
        agent.plugin_rpc.update_device_list.assert_called_once_with(
            agent.context, set(), devices, agent.agent_id, agent.conf.host)

    def test_treat_devices_added_updated_no_device_found(self):
        agent = self.agent
        agent.plugin_rpc = mock.Mock()
        agent.plugin_rpc.get_devices_details_list.return_value = [RPC_DEV1]
        agent.remove_port_binding = mock.Mock()
        agent.eswitch_mgr = mock.Mock()
        agent.eswitch_mgr.device_exists.return_value = False
        devices = {DEV1}
        resync_needed = agent.treat_devices_added_updated(devices)
        self.assertTrue(resync_needed)
        self.assertFalse(agent.plugin_rpc.update_device_up.called)

    def test_update_and_clean_network_ports(self):
        network_id1 = 'network_id1'
        network_id2 = 'network_id2'

        port_id1 = 'port_id1'
        port_id2 = 'port_id2'

        self.agent._update_network_ports(network_id1, port_id1, DEV1)
        self.agent._update_network_ports(network_id1, port_id2, DEV2)

        self.agent._update_network_ports(network_id2, port_id1, DEV1)
        expected = {network_id1: [{'port_id': port_id2, 'device': DEV2}],
                    network_id2: [{'port_id': port_id1, 'device': DEV1}]}
        self.assertEqual(expected, dict(self.agent.network_ports))

        self.assertEqual(port_id1, self.agent._clean_network_ports(DEV1))
        expected = {network_id1: [{'port_id': port_id2, 'device': DEV2}]}
        self.assertEqual(expected, self.agent.network_ports)

        self.assertEqual(port_id2, self.agent._clean_network_ports(DEV2))
        self.assertEqual({}, self.agent.network_ports)

    def test_configurations_has_rp_bandwidth(self):
        rp_bandwidth = {'ens7': {'egress': 10000, 'ingress': 10000}}
        agent = sriov_nic_agent.SriovNicSwitchAgent(
            {}, {}, 0, rp_bandwidth, {}, {})
        self.assertIn(constants.RP_BANDWIDTHS,
                      agent.agent_state['configurations'])

        rp_bandwidths = agent.agent_state['configurations'][
            constants.RP_BANDWIDTHS]
        self.assertEqual(rp_bandwidth['ens7'], rp_bandwidths['ens7'])

    def test_configurations_has_rp_default_inventory(self):
        rp_inventory_values = {
            'allocation_ratio': 1.0,
            'min_unit': 1,
            'step_size': 1,
            'reserved': 0
        }
        agent = sriov_nic_agent.SriovNicSwitchAgent(
            {}, {}, 0, {}, rp_inventory_values, {})
        self.assertIn(constants.RP_INVENTORY_DEFAULTS,
                      agent.agent_state['configurations'])

        rp_inv_defaults = agent.agent_state['configurations'][
            constants.RP_INVENTORY_DEFAULTS]
        self.assertListEqual(
            sorted(list(rp_inventory_values)),
            sorted(list(rp_inv_defaults.keys())))
        for inv_key, inv_value in rp_inventory_values.items():
            self.assertEqual(inv_value,
                             rp_inv_defaults[inv_key])

    def test_process_activated_bindings(self):
        fake_device_info = {
            'current': {DEV1, DEV2},
            'added': {DEV3},
            'removed': set(),
            'updated': set()}
        fake_activated_bindings = {DEV1}
        self.agent.process_activated_bindings(fake_device_info,
                                              fake_activated_bindings)
        self.assertLessEqual(fake_activated_bindings,
                             fake_device_info['added'])


class FakeAgent(object):
    def __init__(self):
        self.updated_devices = set()
        self.activated_bindings = set()
        self.conf = mock.Mock()
        self.conf.host = 'host1'


class TestSriovNicSwitchRpcCallbacks(base.BaseTestCase):

    def setUp(self):
        super(TestSriovNicSwitchRpcCallbacks, self).setUp()
        self.context = object()
        self.agent = FakeAgent()
        sg_agent = object()
        self.sriov_rpc_callback = sriov_nic_agent.SriovNicSwitchRpcCallbacks(
            self.context, self.agent, sg_agent)
        self.device_info = agent_rpc.DeviceInfo(DEVICE_MAC, PCI_SLOT)

    def _create_fake_port(self):
        return {'id': uuidutils.generate_uuid(),
                portbindings.PROFILE: {'pci_slot': PCI_SLOT},
                'mac_address': DEVICE_MAC}

    def _create_fake_bindings(self, fake_port, fake_host):
        return {'port_id': fake_port['id'],
                'host': fake_host}

    def test_port_update_with_pci_slot(self):
        port = self._create_fake_port()
        kwargs = {'context': self.context, 'port': port}
        self.sriov_rpc_callback.port_update(**kwargs)
        self.assertEqual({self.device_info}, self.agent.updated_devices)

    def test_port_update_with_vnic_physical_direct(self):
        port = self._create_fake_port()
        port[portbindings.VNIC_TYPE] = portbindings.VNIC_DIRECT_PHYSICAL
        kwargs = {'context': self.context, 'port': port}
        self.sriov_rpc_callback.port_update(**kwargs)
        self.assertEqual(set(), self.agent.updated_devices)

    def test_port_update_without_pci_slot(self):
        port = self._create_fake_port()
        port[portbindings.PROFILE] = {}
        kwargs = {'context': self.context, 'port': port}
        self.sriov_rpc_callback.port_update(**kwargs)
        self.assertEqual(set(), self.agent.updated_devices)

    def test_network_update(self):
        TEST_NETWORK_ID1 = "n1"
        TEST_NETWORK_ID2 = "n2"
        TEST_PORT_ID1 = 'p1'
        TEST_PORT_ID2 = 'p2'
        network1 = {'id': TEST_NETWORK_ID1}
        port1 = {'id': TEST_PORT_ID1, 'network_id': TEST_NETWORK_ID1}
        port2 = {'id': TEST_PORT_ID2, 'network_id': TEST_NETWORK_ID2}
        self.agent.network_ports = {
                TEST_NETWORK_ID1: [{'port_id': port1['id'],
                                   'device': ('mac1', 'slot1')}],
                TEST_NETWORK_ID2: [{'port_id': port2['id'],
                                   'device': ('mac2', 'slot2')}]}
        kwargs = {'context': self.context, 'network': network1}
        self.sriov_rpc_callback.network_update(**kwargs)
        self.assertEqual(set([('mac1', 'slot1')]), self.agent.updated_devices)

    def test_binding_activate(self):
        fake_port = self._create_fake_port()
        self.agent.get_device_details_from_port_id = mock.Mock()
        self.agent.get_device_details_from_port_id.return_value = {
            'mac_address': fake_port['mac_address'],
            'profile': fake_port[portbindings.PROFILE]
        }
        kwargs = self._create_fake_bindings(fake_port, self.agent.conf.host)
        kwargs['context'] = self.context
        self.sriov_rpc_callback.binding_activate(**kwargs)
        # Assert agent.activated_binding set contains the new binding
        self.assertIn((fake_port['mac_address'],
                       fake_port[portbindings.PROFILE]['pci_slot']),
                      self.agent.activated_bindings)

    def test_binding_activate_no_host(self):
        fake_port = self._create_fake_port()
        kwargs = self._create_fake_bindings(fake_port, 'other-host')
        kwargs['context'] = self.context
        self.sriov_rpc_callback.binding_activate(**kwargs)
        # Assert no bindings were added
        self.assertEqual(set(), self.agent.activated_bindings)

    def test_binding_deactivate(self):
        # binding_deactivate() basically does nothing
        # call it with both the agent's host and other host to cover
        # all code paths
        fake_port = self._create_fake_port()
        kwargs = self._create_fake_bindings(fake_port, self.agent.conf.host)
        kwargs['context'] = self.context
        self.sriov_rpc_callback.binding_deactivate(**kwargs)
        kwargs['host'] = 'other-host'
        self.sriov_rpc_callback.binding_deactivate(**kwargs)


class TestSRIOVAgentExtensionConfig(base.BaseTestCase):
    def setUp(self):
        super(TestSRIOVAgentExtensionConfig, self).setUp()
        l2_ext_manager.register_opts(cfg.CONF)
        # disable setting up periodic state reporting
        cfg.CONF.set_override('report_interval', 0, group='AGENT')
        cfg.CONF.set_override('extensions', ['qos'], group='agent')

    @mock.patch("neutron.plugins.ml2.drivers.mech_sriov.agent.eswitch_manager"
               ".ESwitchManager.get_assigned_devices_info", return_value=[])
    def test_report_loaded_extension(self, *args):
        with mock.patch.object(agent_rpc.PluginReportStateAPI,
                               'report_state') as mock_report_state:
            agent = sriov_nic_agent.SriovNicSwitchAgent({}, {}, 0, {}, {}, {})
            agent._report_state()
            mock_report_state.assert_called_with(
                agent.context, agent.agent_state)
            self.assertEqual(
                ['qos'], agent.agent_state['configurations']['extensions'])


class TestSriovNicAgentConfigParser(base.BaseTestCase):

    def test__validate_rp_in_dev_mappings(self):
        with mock.patch.object(
                cfg.CONF.SRIOV_NIC, 'physical_device_mappings',
                new=[]), \
             mock.patch.object(
                cfg.CONF.SRIOV_NIC, 'resource_provider_bandwidths',
                new=['no_such_dev_in_dev_mappings:1:1']):
            parser = sriov_nic_agent.SriovNicAgentConfigParser()
            self.assertRaises(ValueError, parser.parse)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.db import resource_extend
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.db import rbac_db_models
from neutron.objects import network as network_obj
from neutron.plugins.ml2 import plugin as ml2_plugin
from neutron import quota
from neutron.tests.unit import testlib_api


class NetworkRBACTestCase(testlib_api.SqlTestCase):
    """Base class to test network RBAC policies"""

    def setUp(self):
        super(NetworkRBACTestCase, self).setUp()
        cfg.CONF.set_override('notify_nova_on_port_status_changes', False)
        DB_PLUGIN_KLASS = 'neutron.plugins.ml2.plugin.Ml2Plugin'
        self.setup_coreplugin(DB_PLUGIN_KLASS)
        self.plugin = ml2_plugin.Ml2Plugin()
        self.ctx = context.Context(user_id=None,
                                   tenant_id=None,
                                   is_admin=True,
                                   overwrite=False)
        self.tenant_1 = uuidutils.generate_uuid()
        self.tenant_2 = uuidutils.generate_uuid()
        self.network_id = uuidutils.generate_uuid()
        self.subnet_1_id = uuidutils.generate_uuid()
        self.subnet_2_id = uuidutils.generate_uuid()
        self.port_id = uuidutils.generate_uuid()
        quota_check = mock.patch.object(quota.QuotaEngine, 'quota_limit_check')
        self.mock_quota_check = quota_check.start()

    def _create_network(self, tenant_id, network_id, shared, external=False):
        network = {'tenant_id': tenant_id,
                   'id': network_id,
                   'name': 'test-net',
                   'admin_state_up': True,
                   'shared': shared,
                   extnet_apidef.EXTERNAL: external,
                   'status': constants.NET_STATUS_ACTIVE}
        return self.plugin.create_network(self.ctx, {'network': network})

    def _update_network(self, network_id, network):
        return self.plugin.update_network(self.ctx, network_id,
                                          {'network': network})

    def _create_subnet(self, tenant_id, subnet_id, shared, cidr=None):
        cidr = cidr if cidr else '10.10.10.0/24'
        subnet = {'tenant_id': tenant_id,
                  'id': subnet_id,
                  'name': 'test_sub',
                  'network_id': self.network_id,
                  'ip_version': constants.IP_VERSION_4,
                  'cidr': cidr,
                  'enable_dhcp': False,
                  'gateway_ip': constants.ATTR_NOT_SPECIFIED,
                  'shared': shared,
                  'allocation_pools': constants.ATTR_NOT_SPECIFIED,
                  'dns_nameservers': constants.ATTR_NOT_SPECIFIED,
                  'host_routes': constants.ATTR_NOT_SPECIFIED}
        return self.plugin.create_subnet(self.ctx, {'subnet': subnet})

    def _create_port(self, tenant_id, network_id, port_id):
        port = {'tenant_id': tenant_id,
                'name': 'test_port',
                'id': port_id,
                'network_id': network_id,
                'mac_address': constants.ATTR_NOT_SPECIFIED,
                'admin_state_up': True,
                'status': constants.PORT_STATUS_ACTIVE,
                'device_id': 'test_dev_id',
                'device_owner': constants.DEVICE_OWNER_COMPUTE_PREFIX,
                'fixed_ips': constants.ATTR_NOT_SPECIFIED}
        return self.plugin.create_port(self.ctx, {'port': port})

    def _list_networks(self, ctx):
        return self.plugin.get_networks(ctx)

    def _check_rbac(self, network_id, is_none, external):
        if external:
            action = rbac_db_models.ACCESS_EXTERNAL
        else:
            action = rbac_db_models.ACCESS_SHARED
        rbac = network_obj.NetworkRBAC.get_object(
            self.ctx, object_id=network_id, action=action, target_project='*')
        if is_none:
            self.assertIsNone(rbac)
        else:
            self.assertIsNotNone(rbac)

    def test_network_owner(self):
        tenant_1 = {
            'net-not-shared': (uuidutils.generate_uuid(), False),
            'net-shared': (uuidutils.generate_uuid(), True)}
        tenant_2 = {
            'net-not-shared': (uuidutils.generate_uuid(), False),
            'net-shared': (uuidutils.generate_uuid(), True)}
        for uuid, shared in tenant_1.values():
            self._create_network(self.tenant_1, uuid, shared)
            self._check_rbac(uuid, is_none=(not shared), external=False)
        for uuid, shared in tenant_2.values():
            self._create_network(self.tenant_2, uuid, shared)
            self._check_rbac(uuid, is_none=(not shared), external=False)

        ctx_1 = context.Context(user_id=None,
                                tenant_id=self.tenant_1,
                                is_admin=False,
                                overwrite=False)
        ctx_2 = context.Context(user_id=None,
                                tenant_id=self.tenant_2,
                                is_admin=False,
                                overwrite=False)

        nets_1 = [net['id'] for net in self._list_networks(ctx_1)]
        self.assertEqual(3, len(nets_1))
        self.assertIn(tenant_1['net-shared'][0], nets_1)
        self.assertIn(tenant_1['net-not-shared'][0], nets_1)
        self.assertIn(tenant_2['net-shared'][0], nets_1)

        nets_2 = [net['id'] for net in self._list_networks(ctx_2)]
        self.assertEqual(3, len(nets_2))
        self.assertIn(tenant_2['net-shared'][0], nets_2)
        self.assertIn(tenant_2['net-not-shared'][0], nets_2)
        self.assertIn(tenant_1['net-shared'][0], nets_2)

    def test_create_network_shared(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._check_rbac(self.network_id, is_none=False, external=False)

    def test_create_network_not_shared(self):
        self._create_network(self.tenant_1, self.network_id, False)
        self._check_rbac(self.network_id, is_none=True, external=False)

    def test_create_network_not_shared_external(self):
        with mock.patch.object(resource_extend, 'apply_funcs'):
            self._create_network(self.tenant_1, self.network_id, False,
                                 external=True)
        self._check_rbac(self.network_id, is_none=False, external=True)

    def test_update_network_to_shared(self):
        self._create_network(self.tenant_1, self.network_id, False)
        self._check_rbac(self.network_id, is_none=True, external=False)
        network_data = {'shared': True}
        self._update_network(self.network_id, network_data)
        self._check_rbac(self.network_id, is_none=False, external=False)

    def test_update_network_to_no_shared_no_subnets(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._check_rbac(self.network_id, is_none=False, external=False)

        network_data = {'shared': False}
        self._update_network(self.network_id, network_data)
        self._check_rbac(self.network_id, is_none=True, external=False)

    def test_update_network_shared_to_external(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._check_rbac(self.network_id, is_none=False, external=False)
        self._check_rbac(self.network_id, is_none=True, external=True)

        network_data = {extnet_apidef.EXTERNAL: True}
        self._update_network(self.network_id, network_data)
        self._check_rbac(self.network_id, is_none=False, external=False)
        self._check_rbac(self.network_id, is_none=False, external=True)

    def test_update_network_shared_to_internal(self):
        self._create_network(self.tenant_1, self.network_id, True,
                             external=True)
        self._check_rbac(self.network_id, is_none=False, external=False)
        self._check_rbac(self.network_id, is_none=False, external=True)

        network_data = {extnet_apidef.EXTERNAL: False}
        self._update_network(self.network_id, network_data)
        self._check_rbac(self.network_id, is_none=False, external=False)
        self._check_rbac(self.network_id, is_none=True, external=True)

    def test_update_network_to_no_shared_tenant_subnet(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._check_rbac(self.network_id, is_none=False, external=False)
        self._create_subnet(self.tenant_1, self.subnet_1_id, True)

        network_data = {'shared': False}
        self._update_network(self.network_id, network_data)
        self._check_rbac(self.network_id, is_none=True, external=False)

    def test_update_network_to_no_shared_no_tenant_subnet(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._check_rbac(self.network_id, is_none=False, external=False)
        self._create_subnet(self.tenant_1, self.subnet_1_id, True)
        self._create_subnet(self.tenant_2, self.subnet_2_id, True,
                            cidr='10.10.20/24')

        network_data = {'shared': False}
        self.assertRaises(n_exc.InvalidSharedSetting, self._update_network,
                          self.network_id, network_data)

    def test_ensure_no_port_in_asterisk(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self.plugin.ensure_no_tenant_ports_on_network(
            self.ctx, self.network_id, self.tenant_1, '*')

    def test_ensure_no_port_in_tenant_1(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self.plugin.ensure_no_tenant_ports_on_network(
            self.ctx, self.network_id, self.tenant_1, self.tenant_1)

    def test_ensure_no_port_in_tenant_2(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self.plugin.ensure_no_tenant_ports_on_network(
            self.ctx, self.network_id, self.tenant_1, self.tenant_2)

    def test_ensure_port_tenant_1_in_asterisk(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._create_subnet(self.tenant_1, self.subnet_1_id, True)
        self._create_port(self.tenant_1, self.network_id, self.port_id)
        self.plugin.ensure_no_tenant_ports_on_network(
            self.ctx, self.network_id, self.tenant_1, '*')

    def test_ensure_port_tenant_2_in_asterisk(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._create_subnet(self.tenant_1, self.subnet_1_id, True)
        self._create_port(self.tenant_2, self.network_id, self.port_id)
        self.assertRaises(n_exc.InvalidSharedSetting,
                          self.plugin.ensure_no_tenant_ports_on_network,
                          self.ctx, self.network_id, self.tenant_1, '*')

    def test_ensure_port_tenant_1_in_tenant_1(self):
        self._create_network(self.tenant_1, self.network_id, True)
        self._create_subnet(self.tenant_1, self.subnet_1_id, True)
        self._create_port(self.tenant_1, self.network_id, self.port_id)
        self.plugin.ensure_no_tenant_ports_on_network(
            self.ctx, self.network_id, self.tenant_1, self.tenant_1)

    def test_ensure_no_share_port_tenant_2_in_tenant_1(self):
        self._create_network(self.tenant_1, self.network_id, False)
        self._create_subnet(self.tenant_1, self.subnet_1_id, True)
        self._create_port(self.tenant_2, self.network_id, self.port_id)
        self.plugin.ensure_no_tenant_ports_on_network(
            self.ctx, self.network_id, self.tenant_1, self.tenant_1)

    def test_ensure_no_share_port_tenant_2_in_tenant_2(self):
        self._create_network(self.tenant_1, self.network_id, False)
        self._create_subnet(self.tenant_1, self.subnet_1_id, True)
        self._create_port(self.tenant_2, self.network_id, self.port_id)
        self.assertRaises(n_exc.InvalidSharedSetting,
                          self.plugin.ensure_no_tenant_ports_on_network,
                          self.ctx, self.network_id, self.tenant_1,
                          self.tenant_2)



# Copyright 2021 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

"""increase tag elements from 60 to 255 chars

Revision ID: 8df53b0d2c0e
Revises: 6135a7bd4425
Create Date: 2021-03-29 14:44:35.607053

"""

# revision identifiers, used by Alembic.
revision = '8df53b0d2c0e'
down_revision = '6135a7bd4425'

TABLE = 'tags'


def upgrade():
    op.alter_column(TABLE, 'tag', existing_type=sa.String(60),
                    type_=sa.String(255), existing_nullable=False)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins.ml2 import api
from oslo_log import log


LOG = log.getLogger(__name__)


class LoggerMechanismDriver(api.MechanismDriver):
    """Mechanism driver that logs all calls and parameters made.

    Generally used for testing and debugging.
    """
    def __init__(self, *args, **kwargs):
        super(LoggerMechanismDriver, self).__init__(*args, **kwargs)
        self._supported_extensions = set()

    def initialize(self):
        pass

    def _log_diff_call(self, method_name, context):
        # On the delete the context.current item is the object with
        # all the info and the context.original is the delete object
        # In order to have the output format correct this swap is needed
        if "delete" in method_name:
            og_item = context.current if context.current else {}
            curr_item = context.original if context.original else {}
        else:
            og_item = context.original if context.original else {}
            curr_item = context.current if context.current else {}

        removed_keys = set(og_item.keys()) - set(curr_item.keys())
        added_keys = set(curr_item.keys()) - set(og_item.keys())
        output = "{}:\n".format(method_name)

        for k in curr_item.keys():
            if k in og_item and og_item[k] != curr_item[k]:
                output += "key[{}], {} -> {}\n".format(
                        k, og_item[k], curr_item[k])
        for add_k in added_keys:
            output += "key[{}], None -> {}\n".format(add_k, curr_item[add_k])
        for rem_k in removed_keys:
            output += "key[{}], {} -> None\n".format(rem_k, og_item[rem_k])

        LOG.debug(output)

    def _log_network_call(self, method_name, context):
        LOG.info("%(method)s called with network settings %(current)s "
                 "(original settings %(original)s) and "
                 "network segments %(segments)s",
                 {'method': method_name,
                  'current': context.current,
                  'original': context.original,
                  'segments': context.network_segments})

    def create_network_precommit(self, context):
        self._log_network_call("create_network_precommit", context)
        self._log_diff_call("create_network_precommit", context)

    def create_network_postcommit(self, context):
        self._log_network_call("create_network_postcommit", context)
        self._log_diff_call("create_network_postcommit", context)

    def update_network_precommit(self, context):
        self._log_network_call("update_network_precommit", context)
        self._log_diff_call("update_network_precommit", context)

    def update_network_postcommit(self, context):
        self._log_network_call("update_network_postcommit", context)
        self._log_diff_call("update_network_postcommit", context)

    def delete_network_precommit(self, context):
        self._log_network_call("delete_network_precommit", context)
        self._log_diff_call("delete_network_precommit", context)

    def delete_network_postcommit(self, context):
        self._log_network_call("delete_network_postcommit", context)
        self._log_diff_call("delete_network_postcommit", context)

    def check_vlan_transparency(self, context):
        self._log_network_call("check_vlan_transparency", context)
        self._log_diff_call("check_vlan_transparency", context)
        return True

    def _log_subnet_call(self, method_name, context):
        LOG.info("%(method)s called with subnet settings %(current)s "
                 "(original settings %(original)s)",
                 {'method': method_name,
                  'current': context.current,
                  'original': context.original})

    def create_subnet_precommit(self, context):
        self._log_subnet_call("create_subnet_precommit", context)
        self._log_diff_call("create_subnet_precommit", context)

    def create_subnet_postcommit(self, context):
        self._log_subnet_call("create_subnet_postcommit", context)
        self._log_diff_call("create_subnet_postcommit", context)

    def update_subnet_precommit(self, context):
        self._log_subnet_call("update_subnet_precommit", context)
        self._log_diff_call("update_subnet_precommit", context)

    def update_subnet_postcommit(self, context):
        self._log_subnet_call("update_subnet_postcommit", context)
        self._log_diff_call("update_subnet_postcommit", context)

    def delete_subnet_precommit(self, context):
        self._log_subnet_call("delete_subnet_precommit", context)
        self._log_diff_call("delete_subnet_precommit", context)

    def delete_subnet_postcommit(self, context):
        self._log_subnet_call("delete_subnet_postcommit", context)
        self._log_diff_call("delete_subnet_postcommit", context)

    def _log_port_call(self, method_name, context):
        network_context = context.network
        LOG.info("%(method)s called with port settings %(current)s "
                 "(original settings %(original)s) "
                 "host %(host)s "
                 "(original host %(original_host)s) "
                 "vif type %(vif_type)s "
                 "(original vif type %(original_vif_type)s) "
                 "vif details %(vif_details)s "
                 "(original vif details %(original_vif_details)s) "
                 "binding levels %(levels)s "
                 "(original binding levels %(original_levels)s) "
                 "on network %(network)s "
                 "with segments to bind %(segments_to_bind)s",
                 {'method': method_name,
                  'current': context.current,
                  'original': context.original,
                  'host': context.host,
                  'original_host': context.original_host,
                  'vif_type': context.vif_type,
                  'original_vif_type': context.original_vif_type,
                  'vif_details': context.vif_details,
                  'original_vif_details': context.original_vif_details,
                  'levels': context.binding_levels,
                  'original_levels': context.original_binding_levels,
                  'network': network_context.current,
                  'segments_to_bind': context.segments_to_bind})

    def create_port_precommit(self, context):
        self._log_port_call("create_port_precommit", context)
        self._log_diff_call("create_port_precommit", context)

    def create_port_postcommit(self, context):
        self._log_port_call("create_port_postcommit", context)
        self._log_diff_call("create_port_postcommit", context)

    def update_port_precommit(self, context):
        self._log_port_call("update_port_precommit", context)
        self._log_diff_call("update_port_precommit", context)

    def update_port_postcommit(self, context):
        self._log_port_call("update_port_postcommit", context)
        self._log_diff_call("update_port_postcommit", context)

    def delete_port_precommit(self, context):
        self._log_port_call("delete_port_precommit", context)
        self._log_diff_call("delete_port_precommit", context)

    def delete_port_postcommit(self, context):
        self._log_port_call("delete_port_postcommit", context)
        self._log_diff_call("delete_port_postcommit", context)

    def bind_port(self, context):
        self._log_port_call("bind_port", context)
        self._log_diff_call("bind_port", context)

    def filter_hosts_with_segment_access(
            self, context, segments, candidate_hosts, agent_getter):
        LOG.info("filter_hosts_with_segment_access called with segments "
                 "%(segments)s, candidate hosts %(hosts)s ",
                 {'segments': segments, 'hosts': candidate_hosts})
        return set()

    def supported_extensions(self, extensions):
        if self._supported_extensions:
            return extensions & self._supported_extensions
        return extensions



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import db_base_plugin_common
from neutron.tests import base


class DummyObject(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def to_dict(self):
        return self.kwargs


class ConvertToDictTestCase(base.BaseTestCase):

    @db_base_plugin_common.convert_result_to_dict
    def method_dict(self, fields=None):
        return DummyObject(one=1, two=2, three=3)

    @db_base_plugin_common.convert_result_to_dict
    def method_list(self):
        return [DummyObject(one=1, two=2, three=3)] * 3

    def test_simple_object(self):
        expected = {'one': 1, 'two': 2, 'three': 3}
        observed = self.method_dict()
        self.assertEqual(expected, observed)

    def test_list_of_objects(self):
        expected = [{'one': 1, 'two': 2, 'three': 3}] * 3
        observed = self.method_list()
        self.assertEqual(expected, observed)


class FilterFieldsTestCase(base.BaseTestCase):

    @db_base_plugin_common.filter_fields
    def method_dict(self, fields=None):
        return {'one': 1, 'two': 2, 'three': 3}

    @db_base_plugin_common.filter_fields
    def method_list(self, fields=None):
        return [self.method_dict() for _ in range(3)]

    @db_base_plugin_common.filter_fields
    def method_multiple_arguments(self, not_used, fields=None,
                                  also_not_used=None):
        return {'one': 1, 'two': 2, 'three': 3}

    def test_no_fields(self):
        expected = {'one': 1, 'two': 2, 'three': 3}
        observed = self.method_dict()
        self.assertEqual(expected, observed)

    def test_dict(self):
        expected = {'two': 2}
        observed = self.method_dict(['two'])
        self.assertEqual(expected, observed)

    def test_list(self):
        expected = [{'two': 2}, {'two': 2}, {'two': 2}]
        observed = self.method_list(['two'])
        self.assertEqual(expected, observed)

    def test_multiple_arguments_positional(self):
        expected = {'two': 2}
        observed = self.method_multiple_arguments(list(), ['two'])
        self.assertEqual(expected, observed)

    def test_multiple_arguments_positional_and_keywords(self):
        expected = {'two': 2}
        observed = self.method_multiple_arguments(fields=['two'],
                                                  not_used=None)
        self.assertEqual(expected, observed)

    def test_multiple_arguments_keyword(self):
        expected = {'two': 2}
        observed = self.method_multiple_arguments(list(), fields=['two'])
        self.assertEqual(expected, observed)



# Copyright 2015 HuaWei Technologies.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.services import base as service_base

from neutron.db import models_v2
from neutron.objects import base as base_obj
from neutron.services.timestamp import timestamp_db as ts_db


class TimeStampPlugin(service_base.ServicePluginBase,
                      ts_db.TimeStamp_db_mixin):
    """Implements Neutron Timestamp Service plugin."""

    supported_extension_aliases = ['standard-attr-timestamp']

    __filter_validation_support = True

    def __init__(self):
        super(TimeStampPlugin, self).__init__()
        self.register_db_events()
        # TODO(jlibosva): Move this to register_model_query_hook
        base_obj.register_filter_hook_on_model(
            models_v2.SubnetPool, ts_db.CHANGED_SINCE)

    @classmethod
    def get_plugin_type(cls):
        return 'timestamp'

    def get_plugin_description(self):
        return "Adds timestamps to Neutron resources with standard attributes"



# Copyright 2015 Intel Corporation.
# Copyright 2015 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools
import os

from pyroute2.netlink import exceptions as netlink_exceptions

from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils as linux_utils
from neutron.privileged.agent.linux import ip_lib as priv_ip_lib

# NOTE(toabctl): Don't use /sys/devices/virtual/net here because not all tap
# devices are listed here (i.e. when using Xen)
BRIDGE_FS = "/sys/class/net/"
BRIDGE_INTERFACE_FS = BRIDGE_FS + "%(bridge)s/brif/%(interface)s"
BRIDGE_INTERFACES_FS = BRIDGE_FS + "%s/brif/"
BRIDGE_PORT_FS_FOR_DEVICE = BRIDGE_FS + "%s/brport"
BRIDGE_PATH_FOR_DEVICE = BRIDGE_PORT_FS_FOR_DEVICE + '/bridge'


def catch_exceptions(function):
    """Catch bridge command exceptions

    Returns True if succeeds and False if fails
    """
    @functools.wraps(function)
    def decorated_function(self, *args, **kwargs):
        try:
            function(self, *args, **kwargs)
            return True
        except (RuntimeError, OSError, netlink_exceptions.NetlinkError):
            return False

    return decorated_function


def is_bridged_interface(interface):
    if not interface:
        return False
    else:
        return os.path.exists(BRIDGE_PORT_FS_FOR_DEVICE % interface)


def get_interface_ifindex(interface):
    try:
        with open(os.path.join(BRIDGE_FS, interface, 'ifindex'), 'r') as fh:
            return int(fh.read().strip())
    except (IOError, ValueError):
        pass


def get_bridge_names():
    return os.listdir(BRIDGE_FS)


class BridgeDevice(ip_lib.IPDevice):

    @classmethod
    def addbr(cls, name, namespace=None):
        bridge = cls(name, namespace, 'bridge')
        try:
            bridge.link.create()
        except priv_ip_lib.InterfaceAlreadyExists:
            pass
        return bridge

    @classmethod
    def get_interface_bridge(cls, interface):
        try:
            path = os.readlink(BRIDGE_PATH_FOR_DEVICE % interface)
        except OSError:
            return None
        else:
            name = path.rpartition('/')[-1]
            return cls(name)

    def delbr(self):
        return self.link.delete()

    @catch_exceptions
    def addif(self, interface):
        priv_ip_lib.set_link_bridge_master(interface, self.name,
                                           namespace=self.namespace)

    @catch_exceptions
    def delif(self, interface):
        priv_ip_lib.set_link_bridge_master(interface, None,
                                           namespace=self.namespace)

    @catch_exceptions
    def setfd(self, fd):
        priv_ip_lib.set_link_bridge_forward_delay(self.name, fd,
                                                  namespace=self.namespace)

    @catch_exceptions
    def disable_stp(self):
        priv_ip_lib.set_link_bridge_stp(self.name, 0, namespace=self.namespace)

    @catch_exceptions
    def enable_stp(self):
        priv_ip_lib.set_link_bridge_stp(self.name, 1, namespace=self.namespace)

    def owns_interface(self, interface):
        return os.path.exists(
            BRIDGE_INTERFACE_FS % {'bridge': self.name,
                                   'interface': interface})

    def get_interfaces(self):
        try:
            return os.listdir(BRIDGE_INTERFACES_FS % self.name)
        except OSError:
            return []


class FdbInterface(object):
    """Provide basic functionality to edit the FDB table"""

    @staticmethod
    @catch_exceptions
    def add(mac, dev, dst_ip=None, namespace=None, **kwargs):
        priv_ip_lib.add_bridge_fdb(mac, dev, dst_ip=dst_ip,
                                   namespace=namespace, **kwargs)

    @staticmethod
    @catch_exceptions
    def append(mac, dev, dst_ip=None, namespace=None, **kwargs):
        priv_ip_lib.append_bridge_fdb(mac, dev, dst_ip=dst_ip,
                                      namespace=namespace, **kwargs)

    @staticmethod
    @catch_exceptions
    def replace(mac, dev, dst_ip=None, namespace=None, **kwargs):
        try:
            priv_ip_lib.delete_bridge_fdb(mac, dev, namespace=namespace,
                                          **kwargs)
        except (RuntimeError, OSError, netlink_exceptions.NetlinkError):
            pass
        priv_ip_lib.add_bridge_fdb(mac, dev, dst_ip=dst_ip,
                                   namespace=namespace, **kwargs)

    @staticmethod
    @catch_exceptions
    def delete(mac, dev, dst_ip=None, namespace=None, **kwargs):
        priv_ip_lib.delete_bridge_fdb(mac, dev, dst_ip=dst_ip,
                                      namespace=namespace, **kwargs)

    @staticmethod
    def show(dev=None, namespace=None, **kwargs):
        """List the FDB entries in a namespace

        :parameter dev: device name to filter the query
        :parameter namespace: namespace name
        :returns: a dictionary with the device names and the list of entries
                  per device.
        """

        def find_device_name(ifindex, devices):
            for device in (device for device in devices if
                           device['index'] == ifindex):
                return device['name']

        ret = collections.defaultdict(list)
        fdbs = priv_ip_lib.list_bridge_fdb(namespace=namespace, **kwargs)
        devices = ip_lib.get_devices_info(namespace)
        for fdb in fdbs:
            name = find_device_name(fdb['ifindex'], devices)
            if dev and dev != name:
                continue

            master = find_device_name(linux_utils.get_attr(fdb, 'NDA_MASTER'),
                                      devices)
            fdb_info = {'mac': linux_utils.get_attr(fdb, 'NDA_LLADDR'),
                        'master': master,
                        'vlan': linux_utils.get_attr(fdb, 'NDA_VLAN'),
                        'dst_ip': linux_utils.get_attr(fdb, 'NDA_DST')}
            ret[name].append(fdb_info)

        return ret



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.utils import runtime
from oslo_log import log as logging


LOG = logging.getLogger(__name__)

METERING_NAMESPACE = 'neutron.services.metering_drivers'


def load_metering_driver(plugin, conf):
    """Load metering driver

    :param plugin: the metering plugin
    :param conf: driver configuration object
    :raises SystemExit of 1 if driver cannot be loaded
    """

    try:
        loaded_class = runtime.load_class_by_alias_or_classname(
            METERING_NAMESPACE, conf.driver)
        return loaded_class(plugin, conf)
    except ImportError:
        LOG.error("Error loading metering driver '%s'", conf.driver)
        raise SystemExit(1)



# Copyright 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import shutil

import netaddr
from oslo_log import log as logging

from neutron_lib.api.definitions import portbindings as pbs
from neutron_lib import constants

from neutron.agent.common import async_process
from neutron.agent.linux import ip_lib
from neutron.common import utils
from neutron.tests.common import machine_fixtures
from neutron.tests.common import net_helpers

FULLSTACK_DHCLIENT_SCRIPT = 'fullstack-dhclient-script'
LOG = logging.getLogger(__name__)


class FakeFullstackMachinesList(list):
    """A list of items implementing the FakeFullstackMachine interface."""

    def block_until_all_boot(self):
        for vm in self:
            vm.block_until_boot()

    def block_until_all_dhcp_config_done(self):
        for vm in self:
            vm.block_until_dhcp_config_done()

    def ping_all(self):
        # Generate an iterable of all unique pairs. For example:
        # itertools.permutations(range(3), 2) results in:
        # ((0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1))
        for vm_1, vm_2 in itertools.permutations(self, 2):
            vm_1.block_until_ping(vm_2.ip)

    def ping6_all(self):
        # Generate an iterable of all unique pairs. For example:
        # itertools.permutations(range(3), 2) results in:
        # ((0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1))
        for vm_1, vm_2 in itertools.permutations(self, 2):
            vm_1.block_until_ping(vm_2.ipv6)


class FakeFullstackMachine(machine_fixtures.FakeMachineBase):
    NO_RESOLV_CONF_DHCLIENT_SCRIPT_PATH = (
        shutil.which(FULLSTACK_DHCLIENT_SCRIPT))

    def __init__(self, host, network_id, tenant_id, safe_client,
                 neutron_port=None, bridge_name=None, use_dhcp=False,
                 use_dhcp6=False):
        super(FakeFullstackMachine, self).__init__()
        self.host = host
        self.tenant_id = tenant_id
        self.network_id = network_id
        self.safe_client = safe_client
        self.neutron_port = neutron_port
        self.bridge_name = bridge_name
        self.use_dhcp = use_dhcp
        self.dhclient_async = None
        self.use_dhcp6 = use_dhcp6

    def _setUp(self):
        super(FakeFullstackMachine, self)._setUp()

        self.bridge = self._get_bridge()

        if not self.neutron_port:
            self.neutron_port = self.safe_client.create_port(
                network_id=self.network_id,
                tenant_id=self.tenant_id,
                hostname=self.host.hostname)
        mac_address = self.neutron_port['mac_address']
        hybrid_plug = self.neutron_port[pbs.VIF_DETAILS].get(
            pbs.OVS_HYBRID_PLUG, False)

        self.bind_port_if_needed()
        self.port = self.useFixture(
            net_helpers.PortFixture.get(
                self.bridge, self.namespace, mac_address,
                self.neutron_port['id'], hybrid_plug)).port

        for fixed_ip in self.neutron_port['fixed_ips']:
            self._configure_ipaddress(self.neutron_port['id'], fixed_ip)

    def bind_port_if_needed(self):
        if self.neutron_port[pbs.VIF_TYPE] == pbs.VIF_TYPE_UNBOUND:
            self.safe_client.client.update_port(
                self.neutron_port['id'],
                {'port': {pbs.HOST_ID: self.host.hostname}})
            self.addCleanup(self.safe_client.client.update_port,
                self.neutron_port['id'],
                {'port': {pbs.HOST_ID: ''}})

    def _get_bridge(self):
        if self.bridge_name is None:
            return self.host.get_bridge(self.network_id)
        agent_type = self.host.host_desc.l2_agent_type
        if agent_type == constants.AGENT_TYPE_OVS:
            new_bridge = self.useFixture(
                net_helpers.OVSTrunkBridgeFixture(self.bridge_name)).bridge
        else:
            raise NotImplementedError(
                "Support for %s agent is not implemented." % agent_type)

        return new_bridge

    def _configure_ipaddress(self, port_id, fixed_ip):
        subnet_id = fixed_ip['subnet_id']
        subnet = self.safe_client.client.show_subnet(subnet_id)
        if (netaddr.IPAddress(fixed_ip['ip_address']).version ==
                constants.IP_VERSION_6):
            # v6Address/default_route is auto-configured.
            self._ipv6 = fixed_ip['ip_address']
            self.gateway_ipv6 = subnet['subnet']['gateway_ip']
            if self.use_dhcp6:
                self._configure_ipaddress_via_dhcp(
                    port_id,
                    version=constants.IP_VERSION_6)
        else:
            self._ip = fixed_ip['ip_address']
            prefixlen = netaddr.IPNetwork(subnet['subnet']['cidr']).prefixlen
            self._ip_cidr = '%s/%s' % (self._ip, prefixlen)
            self.gateway_ip = subnet['subnet']['gateway_ip']

            if self.use_dhcp:
                self._configure_ipaddress_via_dhcp(port_id)
            else:
                self._configure_static_ipaddress()

    def _configure_static_ipaddress(self):
        self.port.addr.add(self.ip_cidr)
        if self.gateway_ip:
            net = netaddr.IPNetwork(self.ip_cidr)
            gateway_ip = netaddr.IPAddress(self.gateway_ip)
            if gateway_ip in net:
                net_helpers.set_namespace_gateway(self.port, self.gateway_ip)

    def _configure_ipaddress_via_dhcp(self, port_id,
                                      version=constants.IP_VERSION_4):
        self._start_async_dhclient(port_id, version)
        self.addCleanup(self._stop_async_dhclient)

    def _start_async_dhclient(self, port_id, version=constants.IP_VERSION_4):
        cmd = ["dhclient", '-%s' % version,
               '-lf',
               '%s/%s.lease' % (self.host.neutron_config.temp_dir,
                                port_id),
               '-sf', self.NO_RESOLV_CONF_DHCLIENT_SCRIPT_PATH,
               '--no-pid', '-d', self.port.name]
        self.dhclient_async = async_process.AsyncProcess(
            cmd, run_as_root=True, respawn_interval=5,
            namespace=self.namespace)
        self.dhclient_async.start()

    def _stop_async_dhclient(self):
        if not self.dhclient_async:
            return
        try:
            self.dhclient_async.stop()
            cmd = self.dhclient_async.cmd
            stdout = list(self.dhclient_async.iter_stdout())
            stderr = list(self.dhclient_async.iter_stderr())
            LOG.debug('Stopping async dhclient [%(cmd)s]. stdout: '
                      '[%(stdout)s] - stderr: [%(stderr)s]',
                      {'cmd': cmd, 'stdout': stdout, 'stderr': stderr})
        except async_process.AsyncProcessException:
            # If it was already stopped than we don't care about it
            pass

    @property
    def ipv6(self):
        return self._ipv6

    @property
    def ip(self):
        return self._ip

    @property
    def ip_cidr(self):
        return self._ip_cidr

    def ip_configured(self):
        for port_ip in self.port.addr.list(ip_version=constants.IP_VERSION_4):
            if port_ip.get('cidr') == self.ip_cidr:
                return True
        return False

    def gateway_configured(self):
        gateway_info = self.port.route.get_gateway()
        if not gateway_info:
            return False
        return gateway_info.get('via') == self.gateway_ip

    def block_until_boot(self):
        utils.wait_until_true(
            lambda: (self.safe_client.client.show_port(self.neutron_port['id'])
                     ['port']['status'] == 'ACTIVE'),
            sleep=3)

    def block_until_dhcp_config_done(self):
        utils.wait_until_true(
            lambda: self.ip_configured() and self.gateway_configured(),
            exception=machine_fixtures.FakeMachineException(
                "Address %s or gateway %s not configured properly on "
                "port %s" % (self.ip_cidr, self.gateway_ip, self.port.name)
            )
        )

    def destroy(self, delete_port=False):
        """Destroy this fake machine.

        This should simulate deletion of a vm. It doesn't call cleanUp().
        """
        if delete_port:
            self.safe_client.client.delete_port(self.neutron_port['id'])
        else:
            self.safe_client.client.update_port(self.neutron_port['id'],
                                                {'port': {pbs.HOST_ID: ''}})
        # All associated vlan interfaces are deleted too
        # If VM is connected to Linuxbridge it hasn't got "delete_port" method
        # and it is not necessary to delete tap port connected to this bridge.
        # It is veth pair and will be removed together with VM namespace
        if hasattr(self.bridge, "delete_port"):
            self.bridge.delete_port(self.port.name)

        ip_lib.delete_network_namespace(self.namespace)


class FakeFullstackTrunkMachine(FakeFullstackMachine):
    def __init__(self, trunk, *args, **kwargs):
        super(FakeFullstackTrunkMachine, self).__init__(*args, **kwargs)
        self.trunk = trunk

    def add_vlan_interface(self, mac_address, ip_address, segmentation_id):
        """Add VLAN interface to VM's namespace.

        :param mac_address: MAC address to be set on VLAN interface.
        :param ip_address: The IPNetwork instance containing IP address
                           assigned to the interface.
        :param segmentation_id: VLAN tag added to the interface.
        """
        net_helpers.create_vlan_interface(
            self.namespace, self.port.name, mac_address, ip_address,
            segmentation_id)



# Copyright (c) 2023 Canonical Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import netaddr

from neutron._i18n import _
from neutron.db import l3_db
from neutron.db import l3_gwmode_db
from neutron.objects import ports as port_obj
from neutron.objects import router as l3_obj
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import l3_enable_default_route_bfd
from neutron_lib.api.definitions import l3_enable_default_route_ecmp
from neutron_lib.api.definitions import l3_ext_gw_multihoming
from neutron_lib.api import extensions
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.db import api as db_api
from neutron_lib.db import resource_extend
from neutron_lib.exceptions import l3 as l3_exc
from neutron_lib.exceptions import l3_ext_gw_multihoming as mh_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_config import cfg


def format_gateway_info(gw_port):
    return {
        'network_id': gw_port.network_id,
        'external_fixed_ips': [{
            'ip_address': str(alloc.ip_address),
            'subnet_id': alloc.subnet_id,
        } for alloc in gw_port.fixed_ips]
    }


@resource_extend.has_resource_extenders
class ExtraGatewaysDbOnlyMixin(l3_gwmode_db.L3_NAT_dbonly_mixin):
    """A mixin class to expose a router's extra external gateways."""

    @staticmethod
    @resource_extend.extends([l3_apidef.ROUTERS])
    def _extend_router_dict_extra_gateways(router_res, router_db):
        l3_plugin = directory.get_plugin(plugin_constants.L3)
        if not extensions.is_extension_supported(
                l3_plugin, l3_ext_gw_multihoming.ALIAS):
            return

        external_gateways = []
        for gw_port in [
                rp.port
                for rp in router_db.attached_ports
                if rp.port.device_owner == constants.DEVICE_OWNER_ROUTER_GW]:
            if gw_port.id == router_db.gw_port_id:
                external_gateways.insert(0, format_gateway_info(gw_port))
            else:
                external_gateways.append(format_gateway_info(gw_port))

        router_res[l3_ext_gw_multihoming.EXTERNAL_GATEWAYS] = external_gateways

    @registry.receives(resources.ROUTER, [events.BEFORE_DELETE])
    def _delete_router_remove_external_gateways(self, resource, event,
                                                trigger, payload):
        self._remove_all_gateways(payload.context, payload.resource_id)

    @registry.receives(resources.ROUTER, [events.PRECOMMIT_CREATE])
    def _process_bfd_ecmp_create(self, resource, event, trigger, payload):
        attr_defaults = {
            l3_enable_default_route_ecmp.ENABLE_DEFAULT_ROUTE_ECMP: (
                cfg.CONF.enable_default_route_ecmp),
            l3_enable_default_route_bfd.ENABLE_DEFAULT_ROUTE_BFD: (
                cfg.CONF.enable_default_route_bfd),
        }
        router = payload.latest_state
        router_db = payload.metadata['router_db']
        for attr in attr_defaults.keys():
            value = router.get(attr, attr_defaults[attr])
            if value is not None:
                self.set_extra_attr_value(router_db, attr, value)

    @registry.receives(resources.ROUTER, [events.PRECOMMIT_UPDATE])
    def _process_bfd_ecmp_update(self, resource, event, trigger, payload):
        router = payload.request_body
        router_db = payload.desired_state
        for attr in (l3_enable_default_route_bfd.ENABLE_DEFAULT_ROUTE_BFD,
                     l3_enable_default_route_ecmp.ENABLE_DEFAULT_ROUTE_ECMP):
            value = router.get(attr, None)
            if value is not None:
                self.set_extra_attr_value(router_db, attr, value)

    def _add_external_gateways(
            self, context, router_id, gw_info_list, payload):
        """Add external gateways to a router."""
        added_gateways = []
        if not gw_info_list:
            return added_gateways

        # If a router already has extra gateways specified then they need to
        # be changed via the update API.
        router_db = self._get_router(context, router_id)

        if any(rp.port.device_owner == constants.DEVICE_OWNER_ROUTER_GW
               for rp in router_db.attached_ports):
            # Matching for gateway ports with the same network_id and set of
            # fixed_ips is not needed since an IP allocation would fail in this
            # case. And if fixed IPs don't overlap or are not specified a new
            # port will simply be created.
            extra_gw_info = gw_info_list
        else:
            compat_gw_info = gw_info_list[0]
            compat_payload = copy.deepcopy(payload)
            compat_payload['router'].pop('external_gateways')
            compat_payload['external_gateway_info'] = compat_gw_info

            # Update the first router gateway since we treat it in a special
            # way for compatibility.
            self._update_router_gw_info(context, router_id, compat_gw_info,
                                        compat_payload)
            added_gateways.append(compat_gw_info)

            extra_gw_info = gw_info_list[1:]

        # Go over extra gateway ports and add them to the router.
        for gw_info in extra_gw_info:
            # The ``_validate_gw_info`` and ``_create_extra_gw_port`` methods
            # need an updated version of the router_db object, both as a
            # result of the ``_update_router_gw_info`` call above, and as
            # ports are added.
            router_db = self._get_router(context, router_id)

            # Here we do not need to check for external gateway port IP changes
            # as there are no ports yet.
            ext_ips = gw_info.get('external_fixed_ips', [])

            network_id = self._validate_gw_info(context, gw_info,
                                                ext_ips, router_db)
            self._create_extra_gw_port(context, router_db,
                                       network_id, ext_ips)
            added_gateways.append(gw_info)

        return added_gateways

    def _create_extra_gw_port(self, context, router_db, new_network_id,
                              ext_ips):
        with db_api.CONTEXT_READER.using(context):
            # This function should only be used when we have a compat port id
            # added using the compat API that expects one gateway only.
            if not router_db.gw_port:
                raise mh_exc.UnableToAddExtraGateways(
                    router_id=router_db.id,
                    reason=_('router does not have a compatibility gateway '
                             'port'))

        if not new_network_id:
            return

        subnets = self._core_plugin.get_subnets_by_network(context,
                                                           new_network_id)
        # TODO(dmitriis): publish an events.BEFORE_CREATE event for a new
        # resource type e.g. resources.ROUTER_EXTRA_GATEWAY. Semantically
        # this is a different resource from resources.ROUTER_GATEWAY.
        self._check_for_dup_router_subnets(
            context, router_db,
            subnets,
            constants.DEVICE_OWNER_ROUTER_GW
        )
        self._create_router_gw_port(context, router_db,
                                    new_network_id, ext_ips,
                                    update_gw_port=False)

        # TODO(dmitriis): publish an events.AFTER_CREATE event for a new
        # resource type e.g. resources.ROUTER_EXTRA_GATEWAY. Semantically
        # this is a different resource from resources.ROUTER_GATEWAY.

    def _check_for_dup_router_subnets(self, context, router_db,
                                      new_subnets, new_device_owner):
        """Check for overlapping subnets on different networks.

        This method overrides the one in the base class so the logic will be
        triggered for both the compatibility code that might alter the state
        of a single gateway port in the presence of multiple gateway ports
        (without an override it could result in overlap errors that are not
        relevant with the code base supporting multiple gateway ports attached
        to the same network).

        It is possible to have multiple gateway ports attached to the same
        external network which will cause subnets of ports to overlap but will
        not cause issues with routing. However, attaching multiple gateway
        ports to different networks with overlapping subnet ranges will cause
        routing issues. This function checks for that kind of overlap in
        addition to the compatibility cases such as an overlap between
        internal and external network subnets. This is done using the
        device owner field of a port that is planned to be created by the
        caller: specifically, based on that this argument the method can
        tell if new subnets are meant to be associated with a gateway port
        or an internal port.

        :param context: neutron API request context
        :type context: neutron_lib.context.Context
        :param router_db: The router db object to do a check for.
        :type router: neutron.db.models.l3.Router
        :param new_subnets: A list of new subnets to be added to the router
        :type new_subnets: list[neutron.db.models_v2.Subnet]
        :param new_device_owner: A device owner field for the port that is
                                 going to be created with new subnets.
        """
        router_subnets = []
        ext_subnets = set()
        for p in (rp.port for rp in router_db.attached_ports):
            for ip in p['fixed_ips']:
                existing_port_owner = p.get('device_owner')
                if existing_port_owner == constants.DEVICE_OWNER_ROUTER_GW:
                    ext_subts = self._core_plugin.get_subnets(
                        context.elevated(),
                        filters={'network_id': [p['network_id']]})
                    for sub in ext_subts:
                        router_subnets.append(sub['id'])
                        ext_subnets.add(sub['id'])
                else:
                    router_subnets.append(ip['subnet_id'])
        if not router_subnets:
            return

        # Ignore temporary Prefix Delegation CIDRs
        new_subnets = [s for s in new_subnets
                       if s['cidr'] != constants.PROVISIONAL_IPV6_PD_PREFIX]
        id_filter = {'id': router_subnets}
        subnets = self._core_plugin.get_subnets(context.elevated(),
                                                filters=id_filter)
        for sub in subnets:
            for new_s in new_subnets:
                # Overlapping subnet ranges are a problem if there is an
                # overlap between subnets on different external networks,
                # between internal and external networks or internal networks
                # (including the case where an attempt to add multiple internal
                # ports on the same subnet is made for the same router).
                if not (new_s['id'] in ext_subnets and
                        new_device_owner == constants.DEVICE_OWNER_ROUTER_GW):
                    self._raise_on_subnets_overlap(sub, new_s)

    def _match_requested_gateway_ports(self, context, router_id,
                                       gw_info_list,
                                       no_match_fixed_ip_fatal=True):
        """Match indirect references to gateway ports to the actual ports.

        Returns 3 parameters:

        1. A dictionary which maps matched gateway port ids to
           external_gateway_info dictionaries as they were passed in
        2. A dict with partial matches on fixed ips
        3. A list of gateway info dictionaries for which there aren't any
           existing gateway ports.
        """
        matched_port_ids = {}
        part_matched_port_ids = {}
        nonexistent_port_info = []
        for gw_info in gw_info_list:
            net_id = gw_info['network_id']
            # Find any gateways that might be attached to the same network.
            gw_ports = port_obj.Port.get_ports_by_router_and_network(
                context.elevated(), router_id,
                constants.DEVICE_OWNER_ROUTER_GW, net_id)

            if not gw_ports:
                nonexistent_port_info.append(gw_info)
                continue

            if not gw_info.get('external_fixed_ips'):
                # Allow for one case where external_fixed_ips are not specified
                # in the request but there is only one gateway port attached to
                # particular network on a router - there is no ambiguity about
                # which port do we want to find in this case.
                if len(gw_ports) == 1:
                    gw_port = gw_ports[0]
                    part_matched_port_ids[gw_port['id']] = gw_info
                    continue
                # Matching to specific fixed IPs of gateway ports is done
                # based on the parameters of a request, otherwise it would
                # be unclear which one of the gateway ports to match to.
                raise mh_exc.UnableToMatchGateways(
                    router_id=router_id,
                    reason=_(
                        'multiple gateway ports are attached to the same '
                        'network %s but external_fixed_ips parameter '
                        'is not specified in the request') % net_id)

            for gw_port in gw_ports:
                current_set = set([a.ip_address for a in gw_port['fixed_ips']])
                target_set = set([netaddr.IPAddress(d['ip_address'])
                                  for d in gw_info['external_fixed_ips']])
                # If there is an intersection - it's a partial match.
                if current_set & target_set:
                    part_matched_port_ids[gw_port['id']] = gw_info
                    # It can also be a full match.
                    if current_set == target_set:
                        matched_port_ids[gw_port['id']] = gw_info
                    break
            else:
                if no_match_fixed_ip_fatal:
                    raise mh_exc.UnableToMatchGateways(
                        router_id=router_id,
                        reason=_('could not match a gateway port attached to '
                                 'network %s based on the specified fixed IPs '
                                 '%s') % (net_id,
                                          gw_info['external_fixed_ips']))
                nonexistent_port_info.append(gw_info)
        return matched_port_ids, part_matched_port_ids, nonexistent_port_info

    def _replace_compat_gw_port(self, context, router_db, new_gw_port_id):
        with db_api.CONTEXT_WRITER.using(context):
            router_db['gw_port_id'] = new_gw_port_id

    def _remove_external_gateways(self, context, router_id, gw_info_list,
                                  payload):
        """Remove external gateways from a router."""
        admin_ctx = context.elevated()
        removed_gateways = []
        if not gw_info_list:
            return removed_gateways

        gw_ports = l3_obj.RouterPort.get_gw_port_ids_by_router_id(admin_ctx,
                                                                  router_id)
        if not gw_ports:
            raise mh_exc.UnableToRemoveGateways(
                router_id=router_id,
                reason=_('the router does not have any external gateways'))

        # The `_validate_gw_info` method takes a DB object.
        router_db = self._get_router(context, router_id)

        # Go over extra gateways and validate the specified information.
        for gw_info in gw_info_list:
            ext_ips = gw_info.get(
                'external_fixed_ips', [])
            self._validate_gw_info(context, gw_info, ext_ips, router_db)

        found_gw_port_ids, part_matches, nonexistent_port_info = (
            self._match_requested_gateway_ports(context, router_id,
                                                gw_info_list))
        if nonexistent_port_info:
            raise mh_exc.UnableToMatchGateways(
                router_id=router_id,
                reason=_('could not match gateway port IDs for gateway info '
                         'with networks %s') % (
                             ', '.join(i['network_id']
                                       for i in nonexistent_port_info)))

        # If the compatibility gw_port_id is to be removed, do it after
        # the removal of extra gateway ports but stash up some information.
        compat_gw_port_info = part_matches.pop(router_db['gw_port_id'], None)

        # Actually remove extra gateways first.
        for extra_gw_port_id in part_matches.keys():
            self._delete_extra_gw_port(context, router_id, extra_gw_port_id)
            removed_gateways.append(part_matches[extra_gw_port_id])

        # If the matched gateway port ID includes the compatibility one, handle
        # its removal in a compatible way.
        if compat_gw_port_info:
            # Removal is done by making an empty update using the
            # compatibility interface. This allows reusing pre-removal checks
            # like the FIP presence check.
            self._update_router_gw_info(context, router_id, {}, {})
            removed_gateways.append(compat_gw_port_info)

        # If there are any ports remaining besides the compatibility one
        # and its removal was done, make sure the remaining port becomes
        # the compatibility port. This is not atomic but the extra GW port
        # should not be removed in the process.
        gw_ports = l3_obj.RouterPort.get_gw_port_ids_by_router_id(admin_ctx,
                                                                  router_id)
        if not router_db['gw_port_id'] and len(gw_ports) > 0:
            new_gw_port_id = gw_ports[0]
            new_network_id = port_obj.Port.get_object(
                admin_ctx, id=new_gw_port_id).network_id
            # Replace the gw_port_id on the router object with an existing one.
            self._replace_compat_gw_port(context, router_db, new_gw_port_id)
            # Generate a compatibility payload.
            synthetic_payload = copy.deepcopy(payload)
            synthetic_payload['router'].pop('external_gateways')
            # Here we only need a network_id because the fixed IPs are already
            # assigned and do not need to be changed.
            info = {
                'network_id': new_network_id
            }
            synthetic_payload['router']['external_gateway_info'] = info
            # Finally update the compatibility gateway port.
            self._update_router_gw_info(
                context, router_id, info, synthetic_payload)

        return removed_gateways

    def _router_extra_gw_port_has_floating_ips(self, context, router_id,
                                               gw_port):
        return l3_obj.FloatingIP.count(context, **{
            'router_id': [router_id],
            'floating_network_id': gw_port.network_id,
        })

    def _delete_extra_gw_port(self, context, router_id, gw_port_id):
        admin_ctx = context.elevated()
        gw_port = port_obj.Port.get_object(admin_ctx, id=gw_port_id)
        fip_count = self._router_extra_gw_port_has_floating_ips(context,
                                                                router_id,
                                                                gw_port)
        if fip_count:
            # Check that there are still other gateway ports attached to the
            # same network, otherwise this gateway port cannot be deleted.
            gw_ports = port_obj.Port.get_ports_by_router_and_network(
                admin_ctx, router_id, constants.DEVICE_OWNER_ROUTER_GW,
                gw_port.network_id)
            if len(gw_ports) < 2:
                raise l3_exc.RouterExternalGatewayInUseByFloatingIp(
                    router_id=router_id, net_id=gw_port.network_id)

        # TODO(dmitriis): publish an events.BEFORE_DELETE event for a new
        # resource type e.g. resources.ROUTER_EXTRA_GATEWAY. Semantically this
        # is a different resource from resources.ROUTER_GATEWAY.

        if db_api.is_session_active(admin_ctx.session):
            admin_ctx.GUARD_TRANSACTION = False
        self._core_plugin.delete_port(
            admin_ctx, gw_port_id, l3_port_check=False)

        # TODO(dmitriis): publish an events.AFTER_DELETE event for a new
        # resource type e.g. resources.ROUTER_EXTRA_GATEWAY. Semantically this
        # is a different resource from resources.ROUTER_GATEWAY.

    @db_api.retry_if_session_inactive()
    def add_external_gateways(self, context, router_id, body):
        gateways = body['router'].get('external_gateways',
                                      constants.ATTR_NOT_SPECIFIED)
        if gateways == constants.ATTR_NOT_SPECIFIED:
            return self._get_router(context, router_id)

        external_gateways = self._add_external_gateways(
            context, router_id, gateways, body)

        with db_api.CONTEXT_WRITER.using(context):
            router = self.update_router(
                context, router_id, {
                    'router': {
                        'external_gateways': external_gateways}})
            return {'router': router}

    @db_api.retry_if_session_inactive()
    def remove_external_gateways(self, context, router_id, body):
        gateways = body['router'].get('external_gateways',
                                      constants.ATTR_NOT_SPECIFIED)
        if gateways == constants.ATTR_NOT_SPECIFIED:
            return self._get_router(context, router_id)

        external_gateways = self._remove_external_gateways(
            context, router_id, gateways, body)
        with db_api.CONTEXT_WRITER.using(context):
            router = self.update_router(
                context,
                router_id,
                {'router':
                 {'external_gateways': external_gateways}})
            return {'router': router}

    def _remove_all_gateways(self, context, router_id):
        router_db = self._get_router(context, router_id)
        compat_gw_port_id = router_db['gw_port_id']
        gw_ports = l3_obj.RouterPort.get_gw_port_ids_by_router_id(
            context.elevated(), router_id)
        for gw_port_id in gw_ports:
            if gw_port_id != compat_gw_port_id:
                self._delete_extra_gw_port(context, router_id, gw_port_id)
        if compat_gw_port_id:
            # Remove the compatibility gw port using the compatibility API
            self._update_router_gw_info(context, router_id, {}, {}, router_db)

    def _update_external_gateways(self, context, router_id, gw_info_list,
                                  payload):
        # An empty list means "remove all gateways".
        if not gw_info_list:
            self._remove_all_gateways(context, router_id)
            return {}

        # The `_validate_gw_info` method takes a DB object.
        router_db = self._get_router(context, router_id)

        # Go over extra gateways and validate the specified information.
        for gw_info in gw_info_list:
            ext_ips = gw_info.get(
                'external_fixed_ips', [])
            self._validate_gw_info(context, gw_info, ext_ips, router_db)

        # Find a match for the first gateway in the list.
        found_gw_port_ids, part_matches, nonexistent_port_info = (
            self._match_requested_gateway_ports(
                context, router_id, gw_info_list[:1],
                no_match_fixed_ip_fatal=False))
        # If there is already an existing extra gateway port matching what was
        # requested in the update for the compatibility gw port, simply update
        # the compatibility gw_port_id.
        if part_matches:
            # Replace the gw_port_id on the router object with an existing one.
            self._replace_compat_gw_port(context, router_db,
                                         list(part_matches.keys())[0])

        # The first gw info dict is special as it designates a compat gw. So
        # we simply try to make an update using the compatibility API.
        router_db = self._update_router_gw_info(context, router_id,
                                                gw_info_list[0], {})

        # Find a match for the rest of the gateway list.
        found_gw_port_ids, part_matches, nonexistent_port_info = (
            self._match_requested_gateway_ports(
                context, router_id, gw_info_list[1:],
                no_match_fixed_ip_fatal=False))

        # For partial matches, we need to update the set of fixed IPs for
        # existing ports.
        for gw_port_id, gw_info in part_matches.items():
            # There can be partial matches without any fixed IPs specified,
            # So we check and skip those.
            fixed_ips = gw_info.get('external_fixed_ips')
            if not fixed_ips:
                continue
            self._core_plugin.update_port(
                context.elevated(),
                gw_port_id,
                {'port': {'fixed_ips': fixed_ips}})

        # Identify the set of ports to remove based on the ones that could not
        # be matched based on the supplied external gateways in the request.
        ports_to_remove = [
            format_gateway_info(rp.port)
            for rp in router_db.attached_ports
            if (rp.port.device_owner == constants.DEVICE_OWNER_ROUTER_GW and
                rp.port.id not in found_gw_port_ids.keys() and
                rp.port.id not in part_matches.keys() and
                rp.port.id != router_db.gw_port_id)
        ]
        self._remove_external_gateways(context, router_id, ports_to_remove, {})

        if nonexistent_port_info:
            synthetic_payload = {
                'router': {
                    'external_gateways': nonexistent_port_info}}

            self._add_external_gateways(context, router_id,
                                        nonexistent_port_info,
                                        synthetic_payload)
        return gw_info_list

    @db_api.retry_if_session_inactive()
    def update_external_gateways(self, context, router_id, body):
        gateways = body['router'].get('external_gateways',
                                      constants.ATTR_NOT_SPECIFIED)
        if gateways == constants.ATTR_NOT_SPECIFIED:
            return self._get_router(context, router_id)

        external_gateways = self._update_external_gateways(
            context, router_id, gateways, body)

        with db_api.CONTEXT_WRITER.using(context):
            router = self.update_router(
                context,
                router_id,
                {'router':
                 {'external_gateways': external_gateways}})
            return {'router': router}

    def _update_router_gw_info(self, context, router_id,
                               info, request_body, router=None):
        router_db = super()._update_router_gw_info(context, router_id, info,
                                                   request_body, router)
        # If a compatibility port got removed as a result of a router update
        # (by passing empty info for external_gateway_info) replace it with
        # one of the existing ones.
        gw_ports = l3_obj.RouterPort.get_gw_port_ids_by_router_id(
            context.elevated(), router_id)
        if gw_ports and not router_db['gw_port_id']:
            new_gw_port_id = gw_ports[0]
            self._replace_compat_gw_port(context, router_db, new_gw_port_id)
        return router_db


class ExtraGatewaysMixinDbMixin(ExtraGatewaysDbOnlyMixin,
                                l3_db.L3_NAT_db_mixin):
    pass
